# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t


class ComputeNode:  # cov-slurm
    """The ComputeNode class holds resource information
    about a physical compute node
    """

    def __init__(
        self, node_name: t.Optional[str] = None, node_ppn: t.Optional[int] = None
    ) -> None:
        """Initialize a ComputeNode

        :param node_name: the name of the node
        :param node_ppn: the number of ppn
        """
        self.name: t.Optional[str] = node_name
        self.ppn: t.Optional[int] = node_ppn

    def _is_valid_node(self) -> bool:
        """Check if the node is complete

        Currently, validity is judged by name
        and ppn being not None.

        :returns: True if valid, false otherwise
        """
        if self.name is None:
            return False
        if self.ppn is None:
            return False

        return True


class Partition:  # cov-slurm
    """The partition class holds information about
    a system partition.
    """

    def __init__(self) -> None:
        """Initialize a system partition"""
        self.name: t.Optional[str] = None
        self.min_ppn: t.Optional[int] = None
        self.nodes: t.Set[ComputeNode] = set()

    def _is_valid_partition(self) -> bool:
        """Check if the partition is valid

        Currently, validity is judged by name
        and each ComputeNode being valid

        :returns: True if valid, false otherwise
        """
        if self.name is None:
            return False
        if len(self.nodes) <= 0:
            return False
        for node in self.nodes:
            if not node._is_valid_node():  # pylint: disable=protected-access
                return False

        return True



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
import typing as t

from ..._core.launcher.stepMapping import StepMap
from ...error import AllocationError, LauncherError, SSUnsupportedError
from ...settings import SettingsBase
from .step import Step
from .stepInfo import StepInfo, UnmanagedStepInfo
from .stepMapping import StepMapping
from .taskManager import TaskManager


class Launcher(abc.ABC):  # pragma: no cover
    """Abstract base class of all launchers

    This class provides the interface between the experiment
    controller and the launcher layer. Each launcher supported
    in SmartSim should implement the methods in this class to
    be fully compatible.
    """

    step_mapping: StepMapping
    task_manager: TaskManager

    @abc.abstractmethod
    def create_step(self, name: str, cwd: str, step_settings: SettingsBase) -> Step:
        raise NotImplementedError

    @abc.abstractmethod
    def get_step_update(
        self, step_names: t.List[str]
    ) -> t.List[t.Tuple[str, t.Union[StepInfo, None]]]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_step_nodes(self, step_names: t.List[str]) -> t.List[t.List[str]]:
        raise NotImplementedError

    @abc.abstractmethod
    def run(self, step: Step) -> t.Optional[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self, step_name: str) -> StepInfo:
        raise NotImplementedError

    def add_step_to_mapping_table(self, name: str, step_map: StepMap) -> None:
        """Add a StepMap to the Launcher step mapping table
        making it monitor the step.

        :param name: name of step to be added
        :param step_map: step map of added step
        """
        self.step_mapping[name] = step_map


class WLMLauncher(Launcher):  # cov-wlm
    """The base class for any Launcher that utilizes workload
    manager specific commands. This base class is used to provide
    implemented methods that are alike across all WLM launchers.
    """

    def __init__(self) -> None:
        super().__init__()
        self.task_manager = TaskManager()
        self.step_mapping = StepMapping()

    @property
    @abc.abstractmethod
    def supported_rs(self) -> t.Dict[t.Type[SettingsBase], t.Type[Step]]:
        raise NotImplementedError

    # every launcher utilizing this interface must have a map
    # of supported RunSettings types (see slurmLauncher.py for ex)
    def create_step(
        self, name: str, cwd: str, step_settings: SettingsBase
    ) -> Step:  # cov-wlm
        """Create a WLM job step

        :param name: name of the entity to be launched
        :param cwd: path to launch dir
        :param step_settings: batch or run settings for entity
        :raises SSUnsupportedError: if batch or run settings type isnt supported
        :raises LauncherError: if step creation fails
        :return: step instance
        """
        try:
            step_class = self.supported_rs[type(step_settings)]
        except KeyError:
            raise SSUnsupportedError(
                f"RunSettings type {type(step_settings)} not supported by this launcher"
            ) from None
        try:
            return step_class(name, cwd, step_settings)
        except AllocationError as e:
            raise LauncherError("Step creation failed") from e

    # these methods are implemented in WLM launchers and
    # don't need to be covered here.

    def get_step_nodes(
        self, step_names: t.List[str]
    ) -> t.List[t.List[str]]:  # pragma: no cover
        raise SSUnsupportedError("Node acquisition not supported for this launcher")

    def get_step_update(
        self, step_names: t.List[str]
    ) -> t.List[t.Tuple[str, t.Union[StepInfo, None]]]:  # cov-wlm
        """Get update for a list of job steps

        :param step_names: list of job steps to get updates for
        :return: list of name, job update tuples
        """
        updates: t.List[t.Tuple[str, t.Union[StepInfo, None]]] = []

        # get updates of jobs managed by workload manager (PBS, Slurm, etc)
        # this is primarily batch jobs.
        s_names, step_ids = self.step_mapping.get_ids(step_names, managed=True)
        if len(step_ids) > 0:
            _step_ids = [str(sid) for sid in step_ids]
            s_statuses = self._get_managed_step_update(_step_ids)
            if s_statuses:
                _updates = list(zip(s_names, s_statuses))
                updates.extend(_updates)

        # get updates of unmanaged jobs (Aprun, mpirun, etc)
        # usually jobs started and monitored through the Popen interface
        t_names, task_ids = self.step_mapping.get_ids(step_names, managed=False)
        if len(task_ids) > 0:
            _task_ids = [str(tid) for tid in task_ids]
            t_statuses = self._get_unmanaged_step_update(_task_ids)
            _updates = list(zip(t_names, t_statuses))
            updates.extend(_updates)

        return updates

    def _get_unmanaged_step_update(
        self, task_ids: t.List[str]
    ) -> t.List[UnmanagedStepInfo]:  # cov-wlm
        """Get step updates for Popen managed jobs

        :param task_ids: task id to check
        :return: list of step updates
        """
        updates = []
        for task_id in task_ids:
            stat, return_code, out, err = self.task_manager.get_task_update(task_id)
            update = UnmanagedStepInfo(stat, return_code, out, err)
            updates.append(update)
        return updates

    # pylint: disable-next=no-self-use
    def _get_managed_step_update(
        self,
        step_ids: t.List[str],  # pylint: disable=unused-argument
    ) -> t.List[StepInfo]:  # pragma: no cover
        return []



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterpris
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import dataclasses
import json
import os
import signal
import sys
import time
import typing as t
from pathlib import Path
from types import FrameType

import zmq

from smartsim._core.launcher.dragon.dragonConnector import DragonConnector
from smartsim._core.schemas import (
    DragonHandshakeRequest,
    DragonRequest,
    DragonShutdownRequest,
    request_registry,
)
from smartsim.log import get_logger

"""
Dragon client entrypoint script, used to start a server, send requests to it
and then shut it down.
"""

logger = get_logger("Dragon Client")

SIGNALS = [signal.SIGINT, signal.SIGQUIT, signal.SIGTERM, signal.SIGABRT]


@dataclasses.dataclass
class DragonClientEntrypointArgs:
    submit: Path


def cleanup() -> None:
    """Cleanup resources"""
    logger.debug("Cleaning up")


def parse_requests(request_filepath: Path) -> t.List[DragonRequest]:
    """Parse serialized requests from file

    :param request_filepath: Path to file with serialized requests
    :return: Deserialized requests
    """
    requests: t.List[DragonRequest] = []
    try:
        with open(request_filepath, "r", encoding="utf-8") as request_file:
            req_strings = json.load(fp=request_file)
    except FileNotFoundError as e:
        logger.error(
            "Could not find file with run requests,"
            f"please check whether {request_filepath} exists."
        )
        raise e from None
    except json.JSONDecodeError as e:
        logger.error(f"Could not decode request file {request_filepath}.")
        raise e from None

    requests = [request_registry.from_string(req_str) for req_str in req_strings]

    return requests


def parse_arguments(args: t.List[str]) -> DragonClientEntrypointArgs:
    """Parse arguments used to run entrypoint script

    :param args: Arguments without name of executable
    :raises ValueError: If the request file is not specified
    :return: Parsed arguments
    """
    parser = argparse.ArgumentParser(
        prefix_chars="+",
        description="SmartSim Dragon Client Process, to be used in batch scripts",
    )
    parser.add_argument("+submit", type=str, help="Path to request file", required=True)
    args_ = parser.parse_args(args)

    if not args_.submit:
        raise ValueError("Request file not provided.")

    return DragonClientEntrypointArgs(submit=Path(args_.submit))


def handle_signal(signo: int, _frame: t.Optional[FrameType] = None) -> None:
    """Handle signals sent to this process

    :param signo: Signal number
    :param _frame: Frame, defaults to None
    """
    if not signo:
        logger.info("Received signal with no signo")
    else:
        logger.info(f"Received signal {signo}")
    cleanup()


def register_signal_handlers() -> None:
    """Register signal handlers prior to execution"""
    # make sure to register the cleanup before the start
    # the process so our signaller will be able to stop
    # the server process.
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)


def execute_entrypoint(args: DragonClientEntrypointArgs) -> int:
    """Execute the entrypoint with specified arguments

    :param args: Parsed arguments
    :return: Return code
    """

    try:
        requests = parse_requests(args.submit)
    except Exception:
        logger.error("Dragon client failed to parse request file", exc_info=True)
        return os.EX_OSFILE

    requests.append(DragonShutdownRequest(immediate=False, frontend_shutdown=True))

    connector = DragonConnector()

    for request in requests:
        response = connector.send_request(request)
        if response.error_message is not None:
            logger.error(response.error_message)

    logger.info("Terminated sending requests, waiting for Dragon Server to complete")

    if not connector.can_monitor:
        logger.error(
            "Could not get Dragon Server PID and will not be able to monitor it."
        )
        return os.EX_IOERR

    while True:
        try:
            time.sleep(5)
            connector.send_request(DragonHandshakeRequest())
        except zmq.error.Again:
            logger.debug("Could not reach server, assuming backend has shut down")
            break

    logger.info("Client has finished.")

    return os.EX_OK


def main(args_: t.List[str]) -> int:
    """Execute the dragon client entrypoint as a module"""

    os.environ["PYTHONUNBUFFERED"] = "1"
    logger.info("Dragon client started")

    args = parse_arguments(args_)
    register_signal_handlers()

    try:
        return execute_entrypoint(args)
    except Exception:
        logger.error(
            "An unexpected error occurred in the Dragon client entrypoint",
            exc_info=True,
        )
    finally:
        cleanup()

    return os.EX_SOFTWARE


if __name__ == "__main__":

    sys.exit(main(sys.argv[1:]))



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from ....settings import RunSettings, SettingsBase
from ..launcher import Launcher
from ..step import LocalStep, Step
from ..stepInfo import StepInfo, UnmanagedStepInfo
from ..stepMapping import StepMapping
from ..taskManager import TaskManager


class LocalLauncher(Launcher):
    """Launcher used for spawning proceses on a localhost machine."""

    def __init__(self) -> None:
        self.task_manager = TaskManager()
        self.step_mapping = StepMapping()

    def create_step(self, name: str, cwd: str, step_settings: SettingsBase) -> Step:
        """Create a job step to launch an entity locally

        :return: Step object
        """
        if not isinstance(step_settings, RunSettings):
            raise TypeError(
                "Local Launcher only supports entities with RunSettings, "
                f"not {type(step_settings)}"
            )
        return LocalStep(name, cwd, step_settings)

    def get_step_update(
        self, step_names: t.List[str]
    ) -> t.List[t.Tuple[str, t.Optional[StepInfo]]]:
        """Get status updates of each job step name provided

        :param step_names: list of step_names
        :return: list of tuples for update
        """
        # step ids are process ids of the tasks
        # as there is no WLM intermediary
        updates: t.List[t.Tuple[str, t.Optional[StepInfo]]] = []
        s_names, s_ids = self.step_mapping.get_ids(step_names, managed=False)
        for step_name, step_id in zip(s_names, s_ids):
            status, ret_code, out, err = self.task_manager.get_task_update(str(step_id))
            step_info = UnmanagedStepInfo(status, ret_code, out, err)
            update = (step_name, step_info)
            updates.append(update)
        return updates

    def get_step_nodes(self, step_names: t.List[str]) -> t.List[t.List[str]]:
        """Return the address of nodes assigned to the step

        :param step_names: list of step_names
        :return: list of node addresses

        TODO: Use socket to find the actual Lo address?
        """
        return [["127.0.0.1"] * len(step_names)]

    def run(self, step: Step) -> str:
        """Run a local step created by this launcher. Utilize the shell
           library to execute the command with a Popen. Output and error
           files will be written to the entity path.

        :param step: LocalStep instance to run
        :return: task_id of the newly created step
        """
        if not self.task_manager.actively_monitoring:
            self.task_manager.start()

        out, err = step.get_output_files()
        cmd = step.get_launch_cmd()

        # pylint: disable-next=consider-using-with
        output = open(out, "w+", encoding="utf-8")
        # pylint: disable-next=consider-using-with
        error = open(err, "w+", encoding="utf-8")

        task_id = self.task_manager.start_task(
            cmd, step.cwd, env=step.env, out=output.fileno(), err=error.fileno()
        )

        self.step_mapping.add(step.name, task_id=task_id, managed=False)
        return task_id

    def stop(self, step_name: str) -> UnmanagedStepInfo:
        """Stop a job step

        :param step_name: name of the step to be stopped
        :return: a UnmanagedStepInfo instance
        """
        # step_id is task_id for local. Naming for consistency
        step_id = self.step_mapping[step_name].task_id

        self.task_manager.remove_task(str(step_id))
        _, ret_code, out, err = self.task_manager.get_task_update(str(step_id))
        step_info = UnmanagedStepInfo("Cancelled", ret_code, out, err)
        return step_info

    def __str__(self) -> str:
        return "Local"



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import time
import typing as t
from itertools import product

import redis
from redis.cluster import ClusterNode, RedisCluster
from redis.exceptions import ClusterDownError, RedisClusterException
from smartredis import Client
from smartredis.error import RedisReplyError

from ...entity import DBModel, DBScript
from ...error import SSInternalError
from ...log import get_logger
from ..config import CONFIG
from ..launcher.util.shell import execute_cmd
from .network import get_ip_from_host

logging.getLogger("rediscluster").setLevel(logging.WARNING)
logger = get_logger(__name__)


def create_cluster(hosts: t.List[str], ports: t.List[int]) -> None:  # cov-wlm
    """Connect launched cluster instances.

    Should only be used in the case where cluster initialization
    needs to occur manually which is not often.

    :param hosts: List of hostnames to connect to
    :param ports: List of ports for each hostname
    :raises SmartSimError: if cluster creation fails
    """
    ip_list = []
    for host in hosts:
        ip_address = get_ip_from_host(host)
        for port in ports:
            address = ":".join((ip_address, str(port) + " "))
            ip_list.append(address)

    # call cluster command
    redis_cli = CONFIG.database_cli
    cmd = [redis_cli, "--cluster", "create"]
    cmd += ip_list
    cmd += ["--cluster-replicas", "0", "--cluster-yes"]
    returncode, out, err = execute_cmd(cmd, proc_input="yes", shell=False)

    if returncode != 0:
        logger.error(out)
        logger.error(err)
        raise SSInternalError("Database '--cluster create' command failed")
    logger.debug(out)


def check_cluster_status(
    hosts: t.List[str], ports: t.List[int], trials: int = 10
) -> None:  # cov-wlm
    """Check that a Redis/KeyDB cluster is up and running

    :param hosts: List of hostnames to connect to
    :param ports: List of ports for each hostname
    :param trials: number of attempts to verify cluster status

    :raises SmartSimError: If cluster status cannot be verified
    """
    cluster_nodes = [
        ClusterNode(get_ip_from_host(host), port)
        for host, port in product(hosts, ports)
    ]

    if not cluster_nodes:
        raise SSInternalError(
            "No cluster nodes have been set for database status check."
        )

    logger.debug("Beginning database cluster status check...")
    while trials > 0:
        # wait for cluster to spin up
        time.sleep(5)
        try:
            redis_tester: "RedisCluster[t.Any]" = RedisCluster(
                startup_nodes=cluster_nodes
            )
            redis_tester.set("__test__", "__test__")
            redis_tester.delete("__test__")  # type: ignore
            logger.debug("Cluster status verified")
            return
        except (ClusterDownError, RedisClusterException, redis.RedisError):
            logger.debug("Cluster still spinning up...")
            trials -= 1
    if trials == 0:
        raise SSInternalError("Cluster setup could not be verified")


def db_is_active(hosts: t.List[str], ports: t.List[int], num_shards: int) -> bool:
    """Check if a DB is running

    if the DB is clustered, check cluster status, otherwise
    just ping DB.

    :param hosts: list of hosts
    :param ports: list of ports
    :param num_shards: Number of DB shards
    :return: Whether DB is running
    """
    # if single shard
    if num_shards < 2:
        host = hosts[0]
        port = ports[0]
        try:
            client = redis.Redis(host=host, port=port, db=0)
            if client.ping():
                return True
            return False
        except redis.RedisError:
            return False
    # if a cluster
    else:
        try:
            check_cluster_status(hosts, ports, trials=1)
            return True
        # we expect this to fail if the cluster is not active
        except SSInternalError:
            return False


def set_ml_model(db_model: DBModel, client: Client) -> None:
    logger.debug(f"Adding DBModel named {db_model.name}")

    for device in db_model.devices:
        try:
            if db_model.is_file:
                client.set_model_from_file(
                    name=db_model.name,
                    model_file=str(db_model.file),
                    backend=db_model.backend,
                    device=device,
                    batch_size=db_model.batch_size,
                    min_batch_size=db_model.min_batch_size,
                    min_batch_timeout=db_model.min_batch_timeout,
                    tag=db_model.tag,
                    inputs=db_model.inputs,
                    outputs=db_model.outputs,
                )
            else:
                if db_model.model is None:
                    raise ValueError(f"No model attacted to {db_model.name}")
                client.set_model(
                    name=db_model.name,
                    model=db_model.model,
                    backend=db_model.backend,
                    device=device,
                    batch_size=db_model.batch_size,
                    min_batch_size=db_model.min_batch_size,
                    min_batch_timeout=db_model.min_batch_timeout,
                    tag=db_model.tag,
                    inputs=db_model.inputs,
                    outputs=db_model.outputs,
                )
        except RedisReplyError as error:  # pragma: no cover
            logger.error("Error while setting model on orchestrator.")
            raise error


def set_script(db_script: DBScript, client: Client) -> None:
    logger.debug(f"Adding DBScript named {db_script.name}")

    for device in db_script.devices:
        try:
            if db_script.is_file:
                client.set_script_from_file(
                    name=db_script.name, file=str(db_script.file), device=device
                )
            elif db_script.script:
                if isinstance(db_script.script, str):
                    client.set_script(
                        name=db_script.name, script=db_script.script, device=device
                    )
                elif callable(db_script.script):
                    client.set_function(
                        name=db_script.name, function=db_script.script, device=device
                    )
            else:
                raise ValueError(f"No script or file attached to {db_script.name}")
        except RedisReplyError as error:  # pragma: no cover
            logger.error("Error while setting model on orchestrator.")
            raise error


def shutdown_db_node(host_ip: str, port: int) -> t.Tuple[int, str, str]:  # cov-wlm
    """Send shutdown signal to DB node.

    Should only be used in the case where cluster deallocation
    needs to occur manually. Usually, the SmartSim job manager
    will take care of this automatically.

    :param host_ip: IP of host to connect to
    :param ports: Port to which node is listening
    :return: returncode, output, and error of the process
    """
    redis_cli = CONFIG.database_cli
    cmd = [redis_cli, "-h", host_ip, "-p", str(port), "shutdown"]
    returncode, out, err = execute_cmd(cmd, proc_input="yes", shell=False, timeout=10)

    if returncode != 0:
        logger.error(out)
        err_msg = "Error while shutting down DB node. "
        err_msg += f"Return code: {returncode}, err: {err}"
        logger.error(err_msg)
    elif out:
        logger.debug(out)

    return returncode, out, err



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import functools
import os.path as osp
import pathlib
import sys
import time
import typing as t
from os import makedirs

from smartsim._core.config import CONFIG
from smartsim.error.errors import SmartSimError, UnproxyableStepError

from ....log import get_logger
from ....settings.base import RunSettings, SettingsBase
from ...utils.helpers import encode_cmd, get_base_36_repr
from ..colocated import write_colocated_launch_script

logger = get_logger(__name__)


class Step:
    def __init__(self, name: str, cwd: str, step_settings: SettingsBase) -> None:
        self.name = self._create_unique_name(name)
        self.entity_name = name
        self.cwd = cwd
        self.managed = False
        self.step_settings = step_settings
        self.meta: t.Dict[str, str] = {}

    @property
    def env(self) -> t.Optional[t.Dict[str, str]]:
        """Overridable, read only property for step to specify its environment"""
        return None

    def get_launch_cmd(self) -> t.List[str]:
        raise NotImplementedError

    @staticmethod
    def _create_unique_name(entity_name: str) -> str:
        step_name = entity_name + "-" + get_base_36_repr(time.time_ns())
        return step_name

    @staticmethod
    def _ensure_output_directory_exists(output_dir: str) -> None:
        """Create the directory for the step output if it doesn't exist already"""
        if not osp.exists(output_dir):
            pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    def get_output_files(self) -> t.Tuple[str, str]:
        """Return two paths to error and output files based on metadata directory"""
        try:
            output_dir = self.meta["status_dir"]
        except KeyError as exc:
            raise KeyError("Status directory for this step has not been set.") from exc
        self._ensure_output_directory_exists(output_dir)
        output = osp.join(output_dir, f"{self.entity_name}.out")
        error = osp.join(output_dir, f"{self.entity_name}.err")
        return output, error

    def get_step_file(
        self, ending: str = ".sh", script_name: t.Optional[str] = None
    ) -> str:
        """Get the name for a file/script created by the step class

        Used for Batch scripts, mpmd scripts, etc.
        """
        if script_name:
            script_name = script_name if "." in script_name else script_name + ending
            return osp.join(self.cwd, script_name)
        return osp.join(self.cwd, self.entity_name + ending)

    def get_colocated_launch_script(self) -> str:
        # prep step for colocated launch if specifed in run settings
        script_path = self.get_step_file(
            script_name=osp.join(
                ".smartsim", f"colocated_launcher_{self.entity_name}.sh"
            )
        )
        makedirs(osp.dirname(script_path), exist_ok=True)

        db_settings = {}
        if isinstance(self.step_settings, RunSettings):
            db_settings = self.step_settings.colocated_db_settings or {}

        # db log file causes write contention and kills performance so by
        # default we turn off logging unless user specified debug=True
        if db_settings.get("debug", False):
            db_log_file = self.get_step_file(ending="-db.log")
        else:
            db_log_file = "/dev/null"

        # write the colocated wrapper shell script to the directory for this
        # entity currently being prepped to launch
        write_colocated_launch_script(script_path, db_log_file, db_settings)
        return script_path

    # pylint: disable=no-self-use
    def add_to_batch(self, step: Step) -> None:
        """Add a job step to this batch

        :param step: a job step instance e.g. SrunStep
        """
        raise SmartSimError("add_to_batch not implemented for this step type")


_StepT = t.TypeVar("_StepT", bound=Step)


def proxyable_launch_cmd(
    fn: t.Callable[[_StepT], t.List[str]], /
) -> t.Callable[[_StepT], t.List[str]]:
    @functools.wraps(fn)
    def _get_launch_cmd(self: _StepT) -> t.List[str]:
        """
        Generate a launch command that executes the `JobStep` with the
        indirect launching entrypoint instead of directly. The original
        command is passed to the proxy as a base64 encoded string.

        Steps implementing `get_launch_cmd` and decorated with
        `proxyable_launch_cmd` will generate status updates that can be consumed
        by the telemetry monitor and dashboard"""
        original_cmd_list = fn(self)

        if not CONFIG.telemetry_enabled:
            return original_cmd_list

        if self.managed:
            raise UnproxyableStepError(
                f"Attempting to proxy managed step of type {type(self)} "
                "through the unmanaged step proxy entry point"
            )

        proxy_module = "smartsim._core.entrypoints.indirect"
        entity_type = self.meta["entity_type"]
        status_dir = self.meta["status_dir"]

        logger.debug(f"Encoding command{' '.join(original_cmd_list)}")

        # encode the original cmd to avoid potential collisions and escaping
        # errors when passing it using CLI arguments to the indirect entrypoint
        encoded_cmd = encode_cmd(original_cmd_list)

        # return a new command that executes the proxy and passes
        # the original command as an argument
        return [
            sys.executable,
            "-m",
            proxy_module,
            "+name",
            self.name,
            "+command",
            encoded_cmd,
            "+entity_type",
            entity_type,
            "+telemetry_dir",
            status_dir,
            "+working_dir",
            self.cwd,
        ]

    return _get_launch_cmd



import argparse
import importlib.metadata
import os
import pathlib
import typing as t

from tabulate import tabulate

import smartsim._core._cli.utils as _utils
import smartsim._core.utils.helpers as _helpers
from smartsim._core._install.buildenv import BuildEnv as _BuildEnv

_MISSING_DEP = _helpers.colorize("Not Installed", "red")


def execute(
    _args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    print("\nSmart Python Packages:")
    print(
        tabulate(
            [
                ["SmartSim", _fmt_py_pkg_version("smartsim")],
                ["SmartRedis", _fmt_py_pkg_version("smartredis")],
            ],
            headers=["Name", "Version"],
            tablefmt="fancy_outline",
        ),
        end="\n\n",
    )

    print("Orchestrator Configuration:")
    db_path = _utils.get_db_path()
    db_table = [["Installed", _fmt_installed_db(db_path)]]
    if db_path:
        db_table.append(["Location", str(db_path)])
    print(tabulate(db_table, tablefmt="fancy_outline"), end="\n\n")

    print("Redis AI Configuration:")
    rai_path = _helpers.redis_install_base().parent / "redisai.so"
    rai_table = [["Status", _fmt_installed_redis_ai(rai_path)]]
    if rai_path.is_file():
        rai_table.append(["Location", str(rai_path)])
    print(tabulate(rai_table, tablefmt="fancy_outline"), end="\n\n")

    print("Machine Learning Backends:")
    backends = _helpers.installed_redisai_backends()
    print(
        tabulate(
            [
                [
                    "Tensorflow",
                    _utils.color_bool("tensorflow" in backends),
                    _fmt_py_pkg_version("tensorflow"),
                ],
                [
                    "Torch",
                    _utils.color_bool("torch" in backends),
                    _fmt_py_pkg_version("torch"),
                ],
                [
                    "ONNX",
                    _utils.color_bool("onnxruntime" in backends),
                    _fmt_py_pkg_version("onnx"),
                ],
            ],
            headers=["Name", "Backend Available", "Python Package"],
            tablefmt="fancy_outline",
        ),
        end="\n\n",
    )
    return os.EX_OK


def _fmt_installed_db(db_path: t.Optional[pathlib.Path]) -> str:
    if db_path is None:
        return _MISSING_DEP
    db_name, _ = db_path.name.split("-", 1)
    return _helpers.colorize(db_name.upper(), "green")


def _fmt_installed_redis_ai(rai_path: pathlib.Path) -> str:
    if not rai_path.is_file():
        return _MISSING_DEP
    return _helpers.colorize("Installed", "green")


def _fmt_py_pkg_version(pkg_name: str) -> str:
    try:
        return _helpers.colorize(_BuildEnv.get_py_package_version(pkg_name), "green")
    except importlib.metadata.PackageNotFoundError:
        return _MISSING_DEP



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import typing as t

from ....error import LauncherError
from ....log import get_logger
from ....settings import (
    BsubBatchSettings,
    JsrunSettings,
    MpiexecSettings,
    MpirunSettings,
    OrterunSettings,
    RunSettings,
    SettingsBase,
)
from ....status import SmartSimStatus
from ...config import CONFIG
from ..launcher import WLMLauncher
from ..step import (
    BsubBatchStep,
    JsrunStep,
    LocalStep,
    MpiexecStep,
    MpirunStep,
    OrterunStep,
    Step,
)
from ..stepInfo import LSFBatchStepInfo, LSFJsrunStepInfo, StepInfo
from .lsfCommands import bjobs, bkill, jskill, jslist
from .lsfParser import (
    parse_bjobs_jobid,
    parse_bsub,
    parse_jslist_stepid,
    parse_max_step_id_from_jslist,
)

logger = get_logger(__name__)


class LSFLauncher(WLMLauncher):
    """This class encapsulates the functionality needed
    to launch jobs on systems that use LSF as a workload manager.

    All WLM launchers are capable of launching managed and unmanaged
    jobs. Managed jobs are queried through interaction with with WLM,
    in this case LSF. Unmanaged jobs are held in the TaskManager
    and are managed through references to their launching process ID
    i.e. a psutil.Popen object
    """

    # init in WLMLauncher, launcher.py

    @property
    def supported_rs(self) -> t.Dict[t.Type[SettingsBase], t.Type[Step]]:
        # RunSettings types supported by this launcher
        return {
            JsrunSettings: JsrunStep,
            BsubBatchSettings: BsubBatchStep,
            MpirunSettings: MpirunStep,
            MpiexecSettings: MpiexecStep,
            OrterunSettings: OrterunStep,
            RunSettings: LocalStep,
        }

    def run(self, step: Step) -> t.Optional[str]:
        """Run a job step through LSF

        :param step: a job step instance
        :raises LauncherError: if launch fails
        :return: job step id if job is managed
        """
        if not self.task_manager.actively_monitoring:
            self.task_manager.start()

        cmd_list = step.get_launch_cmd()
        step_id = None
        task_id = None
        if isinstance(step, BsubBatchStep):
            # wait for batch step to submit successfully
            return_code, out, err = self.task_manager.start_and_wait(cmd_list, step.cwd)
            if return_code != 0:
                raise LauncherError(f"Bsub batch submission failed\n {out}\n {err}")
            if out:
                step_id = parse_bsub(out)
                logger.debug(f"Gleaned batch job id: {step_id} for {step.name}")
        elif isinstance(step, JsrunStep):
            self.task_manager.start_task(cmd_list, step.cwd)
            time.sleep(1)
            step_id = self._get_lsf_step_id(step)
            logger.debug(f"Gleaned jsrun step id: {step_id} for {step.name}")
        else:
            # mpirun and local launch don't direct output for us
            out, err = step.get_output_files()

            # pylint: disable-next=consider-using-with
            output = open(out, "w+", encoding="utf-8")
            # pylint: disable-next=consider-using-with
            error = open(err, "w+", encoding="utf-8")
            task_id = self.task_manager.start_task(
                cmd_list, step.cwd, step.env, out=output.fileno(), err=error.fileno()
            )

        self.step_mapping.add(step.name, step_id, task_id, step.managed)
        return step_id

    def stop(self, step_name: str) -> StepInfo:
        """Stop/cancel a job step

        :param step_name: name of the job to stop
        :return: update for job due to cancel
        """
        stepmap = self.step_mapping[step_name]
        if stepmap.managed:
            if stepmap.step_id and "." in stepmap.step_id:
                return_code, _, err = jskill([stepmap.step_id.rpartition(".")[-1]])
            else:
                return_code, _, err = bkill([str(stepmap.step_id)])
            if return_code != 0:
                logger.warning(f"Unable to cancel job step {step_name}\n {err}")
            if stepmap.task_id:
                self.task_manager.remove_task(str(stepmap.task_id))
        else:
            self.task_manager.remove_task(str(stepmap.task_id))

        _, step_info = self.get_step_update([step_name])[0]
        if not step_info:
            raise LauncherError(f"Could not get step_info for job step {step_name}")

        step_info.status = (
            SmartSimStatus.STATUS_CANCELLED
        )  # set status to cancelled instead of failed
        return step_info

    @staticmethod
    def _get_lsf_step_id(step: Step, interval: int = 2) -> str:
        """Get the step_id of last launched step from jslist"""
        time.sleep(interval)
        step_id: t.Optional[str] = None
        trials = CONFIG.wlm_trials
        while trials > 0:
            output, _ = jslist([])
            step_id = parse_max_step_id_from_jslist(output)
            if step_id:
                break
            else:
                time.sleep(interval)
                trials -= 1
        if not step_id:
            raise LauncherError("Could not find id of launched job step")
        if not hasattr(step, "alloc"):
            raise LauncherError("Could not find alloc for launched job step")

        return f"{step.alloc}.{step_id}"

    def _get_managed_step_update(self, step_ids: t.List[str]) -> t.List[StepInfo]:
        """Get step updates for WLM managed jobs

        :param step_ids: list of job step ids
        :return: list of updates for managed jobs
        """
        updates: t.List[StepInfo] = []

        for step_id in step_ids:
            # Batch jobs have integer step id,
            # Jsrun processes have {alloc}.{task_id}
            # Include recently finished jobs
            if "." in str(step_id):
                jsrun_step_id = step_id.rpartition(".")[-1]
                jslist_out, _ = jslist([])
                stat, return_code = parse_jslist_stepid(jslist_out, jsrun_step_id)
                n_rc = int(return_code) if return_code is not None else None
                step_info = LSFJsrunStepInfo(stat, n_rc)
                updates.append(step_info)
            else:
                bjobs_args = ["-a"] + step_ids
                bjobs_out, _ = bjobs(bjobs_args)
                stat = parse_bjobs_jobid(bjobs_out, str(step_id))
                # create LSFBatchStepInfo objects to return
                batch_info = LSFBatchStepInfo(stat, None)
                # account for case where job history is not logged by LSF
                if batch_info.status == SmartSimStatus.STATUS_COMPLETED:
                    batch_info.returncode = 0
                updates.append(batch_info)
        return updates

    def __str__(self) -> str:
        return "LSF"



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import platform
import sys
import typing as t
from pathlib import Path

from tabulate import tabulate

from smartsim._core._cli.scripts.dragon_install import install_dragon
from smartsim._core._cli.utils import SMART_LOGGER_FORMAT, color_bool, pip
from smartsim._core._install import builder
from smartsim._core._install.buildenv import (
    BuildEnv,
    DbEngine,
    SetupError,
    Version_,
    VersionConflictError,
    Versioner,
)
from smartsim._core._install.builder import BuildError, Device
from smartsim._core.config import CONFIG
from smartsim._core.utils.helpers import installed_redisai_backends
from smartsim.error import SSConfigError
from smartsim.log import get_logger

logger = get_logger("Smart", fmt=SMART_LOGGER_FORMAT)

# NOTE: all smartsim modules need full paths as the smart cli
#       may be installed into a different directory.

_TPinningStr = t.Literal["==", "!=", ">=", ">", "<=", "<", "~="]


def check_py_onnx_version(versions: Versioner) -> None:
    """Check Python environment for ONNX installation"""
    _check_packages_in_python_env(
        {
            "onnx": Version_(versions.ONNX),
            "skl2onnx": Version_(versions.REDISAI.skl2onnx),
            "onnxmltools": Version_(versions.REDISAI.onnxmltools),
            "scikit-learn": Version_(getattr(versions.REDISAI, "scikit-learn")),
        },
    )


def check_py_tf_version(versions: Versioner) -> None:
    """Check Python environment for TensorFlow installation"""
    _check_packages_in_python_env({"tensorflow": Version_(versions.TENSORFLOW)})


def check_backends_install() -> bool:
    """Checks if backends have already been installed.
    Logs details on how to proceed forward
    if the RAI_PATH environment variable is set or if
    backends have already been installed.
    """
    rai_path = os.environ.get("RAI_PATH", "")
    installed = installed_redisai_backends()
    msg = ""

    if rai_path and installed:
        msg = (
            f"There is no need to build. backends are already built and "
            f"specified in the environment at 'RAI_PATH': {CONFIG.redisai}"
        )
    elif rai_path and not installed:
        msg = (
            "Before running 'smart build', unset your RAI_PATH environment "
            "variable with 'unset RAI_PATH'."
        )
    elif not rai_path and installed:
        msg = (
            "If you wish to re-run `smart build`, you must first run `smart clean`."
            " The following backend(s) must be removed: " + ", ".join(installed)
        )

    if msg:
        logger.error(msg)

    return not bool(msg)


def build_database(
    build_env: BuildEnv, versions: Versioner, keydb: bool, verbose: bool
) -> None:
    # check database installation
    database_name = "KeyDB" if keydb else "Redis"
    database_builder = builder.DatabaseBuilder(
        build_env(),
        jobs=build_env.JOBS,
        _os=builder.OperatingSystem.from_str(platform.system()),
        architecture=builder.Architecture.from_str(platform.machine()),
        malloc=build_env.MALLOC,
        verbose=verbose,
    )
    if not database_builder.is_built:
        logger.info(
            f"Building {database_name} version {versions.REDIS} "
            f"from {versions.REDIS_URL}"
        )
        database_builder.build_from_git(versions.REDIS_URL, versions.REDIS_BRANCH)
        database_builder.cleanup()
    logger.info(f"{database_name} build complete!")


def build_redis_ai(
    build_env: BuildEnv,
    versions: Versioner,
    device: Device,
    use_torch: bool = True,
    use_tf: bool = True,
    use_onnx: bool = False,
    torch_dir: t.Union[str, Path, None] = None,
    libtf_dir: t.Union[str, Path, None] = None,
    verbose: bool = False,
    torch_with_mkl: bool = True,
) -> None:
    # make sure user isn't trying to do something silly on MacOS
    if build_env.PLATFORM == "darwin" and device == Device.GPU:
        raise BuildError("SmartSim does not support GPU on MacOS")

    # decide which runtimes to build
    print("\nML Backends Requested")
    backends_table = [
        ["PyTorch", versions.TORCH, color_bool(use_torch)],
        ["TensorFlow", versions.TENSORFLOW, color_bool(use_tf)],
        ["ONNX", versions.ONNX, color_bool(use_onnx)],
    ]
    print(tabulate(backends_table, tablefmt="fancy_outline"), end="\n\n")
    print(f"Building for GPU support: {color_bool(device == Device.GPU)}\n")

    if not check_backends_install():
        sys.exit(1)

    # TORCH
    if use_torch and torch_dir:
        torch_dir = Path(torch_dir).resolve()
        if not torch_dir.is_dir():
            raise SetupError(
                f"Could not find requested user Torch installation: {torch_dir}"
            )

    # TF
    if use_tf and libtf_dir:
        libtf_dir = Path(libtf_dir).resolve()
        if not libtf_dir.is_dir():
            raise SetupError(
                f"Could not find requested user TF installation: {libtf_dir}"
            )

    build_env_dict = build_env()

    rai_builder = builder.RedisAIBuilder(
        build_env=build_env_dict,
        jobs=build_env.JOBS,
        _os=builder.OperatingSystem.from_str(platform.system()),
        architecture=builder.Architecture.from_str(platform.machine()),
        torch_dir=str(torch_dir) if torch_dir else "",
        libtf_dir=str(libtf_dir) if libtf_dir else "",
        build_torch=use_torch,
        build_tf=use_tf,
        build_onnx=use_onnx,
        verbose=verbose,
        torch_with_mkl=torch_with_mkl,
    )

    if rai_builder.is_built:
        logger.info("RedisAI installed. Run `smart clean` to remove.")
    else:
        # get the build environment, update with CUDNN env vars
        # if present and building for GPU, otherwise warn the user
        if device == Device.GPU:
            gpu_env = build_env.get_cudnn_env()
            cudnn_env_vars = [
                "CUDNN_LIBRARY",
                "CUDNN_INCLUDE_DIR",
                "CUDNN_INCLUDE_PATH",
                "CUDNN_LIBRARY_PATH",
            ]
            if not gpu_env:
                logger.warning(
                    "CUDNN environment variables not found.\n"
                    f"Looked for {cudnn_env_vars}"
                )
            else:
                build_env_dict.update(gpu_env)
        # update RAI build env with cudnn env vars
        rai_builder.env = build_env_dict

        logger.info(
            f"Building RedisAI version {versions.REDISAI}"
            f" from {versions.REDISAI_URL}"
        )

        # NOTE: have the option to add other builds here in the future
        # like "from_tarball"
        rai_builder.build_from_git(
            versions.REDISAI_URL, versions.REDISAI_BRANCH, device
        )
        logger.info("ML Backends and RedisAI build complete!")


def check_py_torch_version(versions: Versioner, device: Device = Device.CPU) -> None:
    """Check Python environment for TensorFlow installation"""
    if BuildEnv.is_macos():
        if device == Device.GPU:
            raise BuildError("SmartSim does not support GPU on MacOS")
        device_suffix = ""
    else:  # linux
        if device == Device.CPU:
            device_suffix = versions.TORCH_CPU_SUFFIX
        elif device == Device.GPU:
            device_suffix = versions.TORCH_CUDA_SUFFIX
        else:
            raise BuildError("Unrecognized device requested")

    torch_deps = {
        "torch": Version_(f"{versions.TORCH}{device_suffix}"),
        "torchvision": Version_(f"{versions.TORCHVISION}{device_suffix}"),
    }
    missing, conflicts = _assess_python_env(
        torch_deps,
        package_pinning="==",
        validate_installed_version=_create_torch_version_validator(
            with_suffix=device_suffix
        ),
    )

    if len(missing) == len(torch_deps) and not conflicts:
        # All PyTorch deps are not installed and there are no conflicting
        # python packages. We can try to install torch deps into the current env.
        logger.info(
            "Torch version not found in python environment. "
            "Attempting to install via `pip`"
        )
        wheel_device = (
            device.value if device == Device.CPU else device_suffix.replace("+", "")
        )
        pip(
            "install",
            "--extra-index-url",
            f"https://download.pytorch.org/whl/{wheel_device}",
            *(f"{package}=={version}" for package, version in torch_deps.items()),
        )
    elif missing or conflicts:
        logger.warning(_format_incompatible_python_env_message(missing, conflicts))


def _create_torch_version_validator(
    with_suffix: str,
) -> t.Callable[[str, t.Optional[Version_]], bool]:
    def check_torch_version(package: str, version: t.Optional[Version_]) -> bool:
        if not BuildEnv.check_installed(package, version):
            return False
        # Default check only looks at major/minor version numbers,
        # Torch requires we look at the patch as well
        installed = BuildEnv.get_py_package_version(package)
        if with_suffix and with_suffix not in installed.patch:
            raise VersionConflictError(
                package,
                installed,
                version or Version_(f"X.X.X{with_suffix}"),
                msg=(
                    f"{package}=={installed} does not satisfy device "
                    f"suffix requirement: {with_suffix}"
                ),
            )
        return True

    return check_torch_version


def _check_packages_in_python_env(
    packages: t.Mapping[str, t.Optional[Version_]],
    package_pinning: _TPinningStr = "==",
    validate_installed_version: t.Optional[
        t.Callable[[str, t.Optional[Version_]], bool]
    ] = None,
) -> None:
    # TODO: Do not like how the default validation function will always look for
    #       a `==` pinning. Maybe turn `BuildEnv.check_installed` into a factory
    #       that takes a pinning and returns an appropriate validation fn?
    validate_installed_version = validate_installed_version or BuildEnv.check_installed
    missing, conflicts = _assess_python_env(
        packages,
        package_pinning,
        validate_installed_version,
    )

    if missing or conflicts:
        logger.warning(_format_incompatible_python_env_message(missing, conflicts))


def _assess_python_env(
    packages: t.Mapping[str, t.Optional[Version_]],
    package_pinning: _TPinningStr,
    validate_installed_version: t.Callable[[str, t.Optional[Version_]], bool],
) -> t.Tuple[t.List[str], t.List[str]]:
    missing: t.List[str] = []
    conflicts: t.List[str] = []

    for name, version in packages.items():
        spec = f"{name}{package_pinning}{version}" if version else name
        try:
            if not validate_installed_version(name, version):
                # Not installed!
                missing.append(spec)
        except VersionConflictError:
            # Incompatible version found
            conflicts.append(spec)

    return missing, conflicts


def _format_incompatible_python_env_message(
    missing: t.Collection[str], conflicting: t.Collection[str]
) -> str:
    indent = "\n\t"
    fmt_list: t.Callable[[str, t.Collection[str]], str] = lambda n, l: (
        f"{n}:{indent}{indent.join(l)}" if l else ""
    )
    missing_str = fmt_list("Missing", missing)
    conflict_str = fmt_list("Conflicting", conflicting)
    sep = "\n" if missing_str and conflict_str else ""
    return (
        "Python Env Status Warning!\n"
        "Requested Packages are Missing or Conflicting:\n\n"
        f"{missing_str}{sep}{conflict_str}\n\n"
        "Consider installing packages at the requested versions via `pip` or "
        "uninstalling them, installing SmartSim with optional ML dependencies "
        "(`pip install smartsim[ml]`), and running `smart clean && smart build ...`"
    )


def _configure_keydb_build(versions: Versioner) -> None:
    """Configure the redis versions to be used during the build operation"""
    versions.REDIS = Version_("6.2.0")
    versions.REDIS_URL = "https://github.com/EQ-Alpha/KeyDB"
    versions.REDIS_BRANCH = "v6.2.0"

    CONFIG.conf_path = Path(CONFIG.core_path, "config", "keydb.conf")
    if not CONFIG.conf_path.resolve().is_file():
        raise SSConfigError(
            "Database configuration file at REDIS_CONF could not be found"
        )


# pylint: disable-next=too-many-statements
def execute(
    args: argparse.Namespace, _unparsed_args: t.Optional[t.List[str]] = None, /
) -> int:
    verbose = args.v
    keydb = args.keydb
    device = Device(args.device.lower())
    is_dragon_requested = args.dragon
    # torch and tf build by default
    pt = not args.no_pt  # pylint: disable=invalid-name
    tf = not args.no_tf  # pylint: disable=invalid-name
    onnx = args.onnx

    build_env = BuildEnv(checks=True)
    logger.info("Running SmartSim build process...")

    logger.info("Checking requested versions...")
    versions = Versioner()

    logger.debug("Checking for build tools...")

    if verbose:
        logger.info("Build Environment:")
        env = build_env.as_dict()
        env_vars = list(env.keys())
        print(tabulate(env, headers=env_vars, tablefmt="github"), "\n")

    if keydb:
        _configure_keydb_build(versions)

    if verbose:
        db_name: DbEngine = "KEYDB" if keydb else "REDIS"
        logger.info("Version Information:")
        vers = versions.as_dict(db_name=db_name)
        version_names = list(vers.keys())
        print(tabulate(vers, headers=version_names, tablefmt="github"), "\n")

    if is_dragon_requested:
        install_to = CONFIG.core_path / ".dragon"
        return_code = install_dragon(install_to)

        if return_code == 0:
            logger.info("Dragon installation complete")
        elif return_code == 1:
            logger.info("Dragon installation not supported on platform")
        else:
            logger.warning("Dragon installation failed")

    try:
        if not args.only_python_packages:
            # REDIS/KeyDB
            build_database(build_env, versions, keydb, verbose)

            # REDISAI
            build_redis_ai(
                build_env,
                versions,
                device,
                pt,
                tf,
                onnx,
                args.torch_dir,
                args.libtensorflow_dir,
                verbose=verbose,
                torch_with_mkl=args.torch_with_mkl,
            )
    except (SetupError, BuildError) as e:
        logger.error(str(e))
        return os.EX_SOFTWARE

    backends = installed_redisai_backends()
    backends_str = ", ".join(s.capitalize() for s in backends) if backends else "No"
    logger.info(f"{backends_str} backend(s) built")

    try:
        if "torch" in backends:
            check_py_torch_version(versions, device)
        if "tensorflow" in backends:
            check_py_tf_version(versions)
        if "onnxruntime" in backends:
            check_py_onnx_version(versions)
    except (SetupError, BuildError) as e:
        logger.error(str(e))
        return os.EX_SOFTWARE

    logger.info("SmartSim build complete!")
    return os.EX_OK


def configure_parser(parser: argparse.ArgumentParser) -> None:
    """Builds the parser for the command"""
    warn_usage = "(ONLY USE IF NEEDED)"
    parser.add_argument(
        "-v",
        action="store_true",
        default=False,
        help="Enable verbose build process",
    )
    parser.add_argument(
        "--device",
        type=str.lower,
        default=Device.CPU.value,
        choices=[device.value for device in Device],
        help="Device to build ML runtimes for",
    )
    parser.add_argument(
        "--dragon",
        action="store_true",
        default=False,
        help="Install the dragon runtime",
    )
    parser.add_argument(
        "--only_python_packages",
        action="store_true",
        default=False,
        help="Only evaluate the python packages (i.e. skip building backends)",
    )
    parser.add_argument(
        "--no_pt",
        action="store_true",
        default=False,
        help="Do not build PyTorch backend",
    )
    parser.add_argument(
        "--no_tf",
        action="store_true",
        default=False,
        help="Do not build TensorFlow backend",
    )
    parser.add_argument(
        "--onnx",
        action="store_true",
        default=False,
        help="Build ONNX backend (off by default)",
    )
    parser.add_argument(
        "--torch_dir",
        default=None,
        type=str,
        help=f"Path to custom <path>/torch/share/cmake/Torch/ directory {warn_usage}",
    )
    parser.add_argument(
        "--libtensorflow_dir",
        default=None,
        type=str,
        help=f"Path to custom libtensorflow directory {warn_usage}",
    )
    parser.add_argument(
        "--keydb",
        action="store_true",
        default=False,
        help="Build KeyDB instead of Redis",
    )
    parser.add_argument(
        "--no_torch_with_mkl",
        dest="torch_with_mkl",
        action="store_false",
        help="Do not build Torch with Intel MKL",
    )



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import json
import time
import typing as t
from pathlib import Path

import smartsim._core._cli.utils as _utils
import smartsim.log

if t.TYPE_CHECKING:
    from smartsim._core.control.manifest import LaunchedManifest as _Manifest
    from smartsim.database.orchestrator import Orchestrator
    from smartsim.entity import DBNode, Ensemble, Model
    from smartsim.entity.dbobject import DBModel, DBScript
    from smartsim.settings.base import BatchSettings, RunSettings


TStepLaunchMetaData = t.Tuple[
    t.Optional[str], t.Optional[str], t.Optional[bool], str, str, Path
]

MANIFEST_FILENAME: t.Final[str] = "manifest.json"

_LOGGER = smartsim.log.get_logger(__name__)


def save_launch_manifest(manifest: _Manifest[TStepLaunchMetaData]) -> None:
    manifest.metadata.run_telemetry_subdirectory.mkdir(parents=True, exist_ok=True)
    exp_out, exp_err = smartsim.log.get_exp_log_paths()

    new_run = {
        "run_id": manifest.metadata.run_id,
        "timestamp": int(time.time_ns()),
        "model": [
            _dictify_model(model, *telemetry_metadata)
            for model, telemetry_metadata in manifest.models
        ],
        "orchestrator": [
            _dictify_db(db, nodes_info) for db, nodes_info in manifest.databases
        ],
        "ensemble": [
            _dictify_ensemble(ens, member_info)
            for ens, member_info in manifest.ensembles
        ],
    }
    try:
        with open(manifest.metadata.manifest_file_path, "r", encoding="utf-8") as file:
            manifest_dict = json.load(file)
    except (FileNotFoundError, json.JSONDecodeError):
        manifest_dict = {
            "schema info": {
                "schema_name": "entity manifest",
                "version": "0.0.4",
            },
            "experiment": {
                "name": manifest.metadata.exp_name,
                "path": manifest.metadata.exp_path,
                "launcher": manifest.metadata.launcher_name,
                "out_file": str(exp_out),
                "err_file": str(exp_err),
            },
            "runs": [new_run],
        }
    else:
        manifest_dict["runs"].append(new_run)
    finally:
        with open(manifest.metadata.manifest_file_path, "w", encoding="utf-8") as file:
            json.dump(manifest_dict, file, indent=2)


def _dictify_model(
    model: Model,
    step_id: t.Optional[str],
    task_id: t.Optional[str],
    managed: t.Optional[bool],
    out_file: str,
    err_file: str,
    telemetry_data_path: Path,
) -> t.Dict[str, t.Any]:
    colo_settings = (model.run_settings.colocated_db_settings or {}).copy()
    db_scripts = t.cast("t.List[DBScript]", colo_settings.pop("db_scripts", []))
    db_models = t.cast("t.List[DBModel]", colo_settings.pop("db_models", []))
    return {
        "name": model.name,
        "path": model.path,
        "exe_args": model.run_settings.exe_args,
        "run_settings": _dictify_run_settings(model.run_settings),
        "batch_settings": (
            _dictify_batch_settings(model.batch_settings)
            if model.batch_settings
            else {}
        ),
        "params": model.params,
        "files": (
            {
                "Symlink": model.files.link,
                "Configure": model.files.tagged,
                "Copy": model.files.copy,
            }
            if model.files
            else {
                "Symlink": [],
                "Configure": [],
                "Copy": [],
            }
        ),
        "colocated_db": (
            {
                "settings": colo_settings,
                "scripts": [
                    {
                        script.name: {
                            "backend": "TORCH",
                            "device": script.device,
                        }
                    }
                    for script in db_scripts
                ],
                "models": [
                    {
                        model.name: {
                            "backend": model.backend,
                            "device": model.device,
                        }
                    }
                    for model in db_models
                ],
            }
            if colo_settings
            else {}
        ),
        "telemetry_metadata": {
            "status_dir": str(telemetry_data_path),
            "step_id": step_id,
            "task_id": task_id,
            "managed": managed,
        },
        "out_file": out_file,
        "err_file": err_file,
    }


def _dictify_ensemble(
    ens: Ensemble,
    members: t.Sequence[t.Tuple[Model, TStepLaunchMetaData]],
) -> t.Dict[str, t.Any]:
    return {
        "name": ens.name,
        "params": ens.params,
        "batch_settings": (
            _dictify_batch_settings(ens.batch_settings)
            # FIXME: Typehint here is wrong, ``ens.batch_settings`` can
            # also be an empty dict for no discernible reason...
            if ens.batch_settings
            else {}
        ),
        "models": [
            _dictify_model(model, *launching_metadata)
            for model, launching_metadata in members
        ],
    }


def _dictify_run_settings(run_settings: RunSettings) -> t.Dict[str, t.Any]:
    # TODO: remove this downcast
    if hasattr(run_settings, "mpmd") and run_settings.mpmd:
        _LOGGER.warning(
            "SmartSim currently cannot properly serialize all information in "
            "MPMD run settings"
        )
    return {
        "exe": run_settings.exe,
        # TODO: We should try to move this back
        # "exe_args": run_settings.exe_args,
        "run_command": run_settings.run_command,
        "run_args": run_settings.run_args,
        # TODO: We currently do not have a way to represent MPMD commands!
        #       Maybe add a ``"mpmd"`` key here that is a
        #       ``list[TDictifiedRunSettings]``?
    }


def _dictify_batch_settings(batch_settings: BatchSettings) -> t.Dict[str, t.Any]:
    return {
        "batch_command": batch_settings.batch_cmd,
        "batch_args": batch_settings.batch_args,
    }


def _dictify_db(
    db: Orchestrator,
    nodes: t.Sequence[t.Tuple[DBNode, TStepLaunchMetaData]],
) -> t.Dict[str, t.Any]:
    db_path = _utils.get_db_path()
    if db_path:
        db_type, _ = db_path.name.split("-", 1)
    else:
        db_type = "Unknown"

    return {
        "name": db.name,
        "type": db_type,
        "interface": db._interfaces,  # pylint: disable=protected-access
        "shards": [
            {
                **shard.to_dict(),
                "conf_file": shard.cluster_conf_file,
                "out_file": out_file,
                "err_file": err_file,
                "memory_file": (
                    str(status_dir / "memory.csv") if db.telemetry.is_enabled else ""
                ),
                "client_file": (
                    str(status_dir / "client.csv") if db.telemetry.is_enabled else ""
                ),
                "client_count_file": (
                    str(status_dir / "client_count.csv")
                    if db.telemetry.is_enabled
                    else ""
                ),
                "telemetry_metadata": {
                    "status_dir": str(status_dir),
                    "step_id": step_id,
                    "task_id": task_id,
                    "managed": managed,
                },
            }
            for dbnode, (
                step_id,
                task_id,
                managed,
                out_file,
                err_file,
                status_dir,
            ) in nodes
            for shard in dbnode.get_launched_shard_info()
        ],
    }



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import typing as t

from ....settings import Singularity
from ....settings.base import RunSettings
from .step import Step, proxyable_launch_cmd


class LocalStep(Step):
    def __init__(self, name: str, cwd: str, run_settings: RunSettings):
        super().__init__(name, cwd, run_settings)
        self.run_settings = run_settings
        self._env = self._set_env()

    @property
    def env(self) -> t.Dict[str, str]:
        return self._env

    @proxyable_launch_cmd
    def get_launch_cmd(self) -> t.List[str]:
        cmd = []

        # Add run command and args if user specified
        # default is no run command for local job steps
        if self.run_settings.run_command:
            cmd.append(self.run_settings.run_command)
            run_args = self.run_settings.format_run_args()
            cmd.extend(run_args)

        if self.run_settings.colocated_db_settings:
            # Replace the command with the entrypoint wrapper script
            if not (bash := shutil.which("bash")):
                raise RuntimeError("Unable to locate bash interpreter")

            launch_script_path = self.get_colocated_launch_script()
            cmd.extend([bash, launch_script_path])

        container = self.run_settings.container
        if container and isinstance(container, Singularity):
            # pylint: disable-next=protected-access
            cmd += container._container_cmds(self.cwd)

        # build executable
        cmd.extend(self.run_settings.exe)
        if self.run_settings.exe_args:
            cmd.extend(self.run_settings.exe_args)
        return cmd

    def _set_env(self) -> t.Dict[str, str]:
        env = os.environ.copy()
        if self.run_settings.env_vars:
            for k, v in self.run_settings.env_vars.items():
                env[k] = v or ""
        return env



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pathlib
import time
import typing as t
from dataclasses import dataclass

from ...entity import EntitySequence, SmartSimEntity
from ...status import SmartSimStatus


@dataclass(frozen=True)
class _JobKey:
    """A helper class for creating unique lookup keys within the telemetry
    monitor. These keys are not guaranteed to be unique across experiments,
    only within an experiment (due to process ID re-use by the OS)"""

    step_id: str
    """The process id of an unmanaged task"""
    task_id: str
    """The task id of a managed task"""


class JobEntity:
    """An entity containing run-time SmartSimEntity metadata. The run-time metadata
    is required to perform telemetry collection. The `JobEntity` satisfies the core
    API necessary to use a `JobManager` to manage retrieval of managed step updates.
    """

    def __init__(self) -> None:
        self.name: str = ""
        """The entity name"""
        self.path: str = ""
        """The root path for entity output files"""
        self.step_id: str = ""
        """The process id of an unmanaged task"""
        self.task_id: str = ""
        """The task id of a managed task"""
        self.type: str = ""
        """The type of the associated `SmartSimEntity`"""
        self.timestamp: int = 0
        """The timestamp when the entity was created"""
        self.status_dir: str = ""
        """The path configured by the experiment for the entities telemetry output"""
        self.telemetry_on: bool = False
        """"Flag indicating if optional telemetry is enabled for the entity"""
        self.collectors: t.Dict[str, str] = {}
        """Mapping of collectors enabled for the entity"""
        self.config: t.Dict[str, str] = {}
        """Telemetry configuration supplied by the experiment"""
        self._is_complete: bool = False
        """Flag indicating if the entity has completed execution"""

    @property
    def is_db(self) -> bool:
        """Returns `True` if the entity represents a database or database shard"""
        return self.type in ["orchestrator", "dbnode"]

    @property
    def is_managed(self) -> bool:
        """Returns `True` if the entity is managed by a workload manager"""
        return bool(self.step_id)

    @property
    def key(self) -> _JobKey:
        """Return a `_JobKey` that identifies an entity.
        NOTE: not guaranteed to be unique over time due to reused process IDs"""
        return _JobKey(self.step_id, self.task_id)

    @property
    def is_complete(self) -> bool:
        """Returns `True` if the entity has completed execution"""
        return self._is_complete

    def check_completion_status(self) -> None:
        """Check for telemetry outputs indicating the entity has completed
        TODO: determine correct location to avoid exposing telemetry
        implementation details into `JobEntity`
        """
        # avoid touching file-system if not necessary
        if self._is_complete:
            return

        # status telemetry is tracked in JSON files written to disk. look
        # for a corresponding `stop` event in the entity status directory
        state_file = pathlib.Path(self.status_dir) / "stop.json"
        if state_file.exists():
            self._is_complete = True

    @staticmethod
    def _map_db_metadata(entity_dict: t.Dict[str, t.Any], entity: "JobEntity") -> None:
        """Map DB-specific properties from a runtime manifest onto a `JobEntity`

        :param entity_dict: The raw dictionary deserialized from manifest JSON
        :param entity: The entity instance to modify
        """
        if entity.is_db:
            # add collectors if they're configured to be enabled in the manifest
            entity.collectors = {
                "client": entity_dict.get("client_file", ""),
                "client_count": entity_dict.get("client_count_file", ""),
                "memory": entity_dict.get("memory_file", ""),
            }

            entity.telemetry_on = any(entity.collectors.values())
            entity.config["host"] = entity_dict.get("hostname", "")
            entity.config["port"] = entity_dict.get("port", "")

    @staticmethod
    def _map_standard_metadata(
        entity_type: str,
        entity_dict: t.Dict[str, t.Any],
        entity: "JobEntity",
        exp_dir: str,
        raw_experiment: t.Dict[str, t.Any],
    ) -> None:
        """Map universal properties from a runtime manifest onto a `JobEntity`

        :param entity_type: The type of the associated `SmartSimEntity`
        :param entity_dict: The raw dictionary deserialized from manifest JSON
        :param entity: The entity instance to modify
        :param exp_dir: The path to the experiment working directory
        :param raw_experiment: The raw experiment dictionary deserialized from
        manifest JSON
        """
        metadata = entity_dict["telemetry_metadata"]
        status_dir = pathlib.Path(metadata.get("status_dir"))
        is_dragon = raw_experiment["launcher"].lower() == "dragon"

        # all entities contain shared properties that identify the task
        entity.type = entity_type
        entity.name = (
            entity_dict["name"]
            if not is_dragon
            else entity_dict["telemetry_metadata"]["step_id"]
        )
        entity.step_id = str(metadata.get("step_id") or "")
        entity.task_id = str(metadata.get("task_id") or "")
        entity.timestamp = int(entity_dict.get("timestamp", "0"))
        entity.path = str(exp_dir)
        entity.status_dir = str(status_dir)

    @classmethod
    def from_manifest(
        cls,
        entity_type: str,
        entity_dict: t.Dict[str, t.Any],
        exp_dir: str,
        raw_experiment: t.Dict[str, t.Any],
    ) -> "JobEntity":
        """Instantiate a `JobEntity` from the dictionary deserialized from manifest JSON

        :param entity_type: The type of the associated `SmartSimEntity`
        :param entity_dict: The raw dictionary deserialized from manifest JSON
        :param exp_dir: The path to the experiment working directory
        :param raw_experiment: raw experiment deserialized from manifest JSON
        """
        entity = JobEntity()

        cls._map_standard_metadata(
            entity_type, entity_dict, entity, exp_dir, raw_experiment
        )
        cls._map_db_metadata(entity_dict, entity)

        return entity


class Job:
    """Keep track of various information for the controller.
    In doing so, continuously add various fields of information
    that is queryable by the user through interface methods in
    the controller class.
    """

    def __init__(
        self,
        job_name: str,
        job_id: t.Optional[str],
        entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity], JobEntity],
        launcher: str,
        is_task: bool,
    ) -> None:
        """Initialize a Job.

        :param job_name: Name of the job step
        :param job_id: The id associated with the job
        :param entity: The SmartSim entity(list) associated with the job
        :param launcher: Launcher job was started with
        :param is_task: process monitored by TaskManager (True) or the WLM (True)
        """
        self.name = job_name
        self.jid = job_id
        self.entity = entity
        self.status = SmartSimStatus.STATUS_NEW
        # status before smartsim status mapping is applied
        self.raw_status: t.Optional[str] = None
        self.returncode: t.Optional[int] = None
        # output is only populated if it's system related (e.g. cmd failed immediately)
        self.output: t.Optional[str] = None
        self.error: t.Optional[str] = None  # same as output
        self.hosts: t.List[str] = []  # currently only used for DB jobs
        self.launched_with = launcher
        self.is_task = is_task
        self.start_time = time.time()
        self.history = History()

    @property
    def ename(self) -> str:
        """Return the name of the entity this job was created from"""
        return self.entity.name

    def set_status(
        self,
        new_status: SmartSimStatus,
        raw_status: str,
        returncode: t.Optional[int],
        error: t.Optional[str] = None,
        output: t.Optional[str] = None,
    ) -> None:
        """Set the status  of a job.

        :param new_status: The new status of the job
        :param raw_status: The raw status of the launcher
        :param returncode: The return code for the job
        :param error: Content produced by stderr
        :param output: Content produced by stdout
        """
        self.status = new_status
        self.raw_status = raw_status
        self.returncode = returncode
        self.error = error
        self.output = output

    @property
    def elapsed(self) -> float:
        return time.time() - self.start_time

    def record_history(self) -> None:
        """Record the launching history of a job."""
        self.history.record(self.jid, self.status, self.returncode, self.elapsed)

    def reset(
        self, new_job_name: str, new_job_id: t.Optional[str], is_task: bool
    ) -> None:
        """Reset the job in order to be able to restart it.

        :param new_job_name: name of the new job step
        :param new_job_id: new job id to launch under
        :param is_task: process monitored by TaskManager (True) or the WLM (True)
        """
        self.name = new_job_name
        self.jid = new_job_id
        self.status = SmartSimStatus.STATUS_NEW
        self.returncode = None
        self.output = None
        self.error = None
        self.hosts = []
        self.is_task = is_task
        self.start_time = time.time()
        self.history.new_run()

    def error_report(self) -> str:
        """A descriptive error report based on job fields

        :return: error report for display in terminal
        """
        warning = f"{self.ename} failed. See below for details \n"
        if self.error:
            warning += (
                f"{self.entity.type} {self.ename} produced the following error \n"
            )
            warning += f"Error: {self.error} \n"
        if self.output:
            warning += f"Output: {self.output} \n"
        warning += f"Job status at failure: {self.status} \n"
        warning += f"Launcher status at failure: {self.raw_status} \n"
        warning += f"Job returncode: {self.returncode} \n"
        warning += f"Error and output file located at: {self.entity.path}"
        return warning

    def __str__(self) -> str:
        """Return user-readable string of the Job

        :returns: A user-readable string of the Job
        """
        if self.jid:
            job = "{}({}): {}"
            return job.format(self.ename, self.jid, self.status)

        job = "{}: {}"
        return job.format(self.ename, self.status)


class History:
    """History of a job instance. Holds various attributes based
    on the previous launches of a job.
    """

    def __init__(self, runs: int = 0) -> None:
        """Init a history object for a job

        :param runs: number of runs so far
        """
        self.runs = runs
        self.jids: t.Dict[int, t.Optional[str]] = {}
        self.statuses: t.Dict[int, SmartSimStatus] = {}
        self.returns: t.Dict[int, t.Optional[int]] = {}
        self.job_times: t.Dict[int, float] = {}

    def record(
        self,
        job_id: t.Optional[str],
        status: SmartSimStatus,
        returncode: t.Optional[int],
        job_time: float,
    ) -> None:
        """record the history of a job"""
        self.jids[self.runs] = job_id
        self.statuses[self.runs] = status
        self.returns[self.runs] = returncode
        self.job_times[self.runs] = job_time

    def new_run(self) -> None:
        """increment run total"""
        self.runs += 1



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
import json
import logging
import os
import pathlib
import threading
import typing as t

from watchdog.events import (
    FileSystemEvent,
    LoggingEventHandler,
    PatternMatchingEventHandler,
)
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver

from smartsim._core.config import CONFIG
from smartsim._core.control.job import JobEntity, _JobKey
from smartsim._core.control.jobmanager import JobManager
from smartsim._core.launcher.dragon.dragonLauncher import DragonLauncher
from smartsim._core.launcher.launcher import Launcher
from smartsim._core.launcher.local.local import LocalLauncher
from smartsim._core.launcher.lsf.lsfLauncher import LSFLauncher
from smartsim._core.launcher.pbs.pbsLauncher import PBSLauncher
from smartsim._core.launcher.slurm.slurmLauncher import SlurmLauncher
from smartsim._core.launcher.stepInfo import StepInfo
from smartsim._core.utils.helpers import get_ts_ms
from smartsim._core.utils.serialize import MANIFEST_FILENAME
from smartsim._core.utils.telemetry.collector import CollectorManager
from smartsim._core.utils.telemetry.manifest import Run, RuntimeManifest
from smartsim._core.utils.telemetry.util import map_return_code, write_event
from smartsim.error.errors import SmartSimError
from smartsim.status import TERMINAL_STATUSES

logger = logging.getLogger("TelemetryMonitor")


class ManifestEventHandler(PatternMatchingEventHandler):
    """The ManifestEventHandler monitors an experiment and updates a
    datastore as needed. This event handler is triggered by changes to
    the experiment manifest written to physical disk by a driver.

    It also contains an event loop. The loop checks experiment entities for updates
    at each timestep and executes a configurable set of metrics collectors."""

    def __init__(
        self,
        pattern: str,
        ignore_patterns: t.Optional[t.List[str]] = None,
        ignore_directories: bool = True,
        case_sensitive: bool = False,
        timeout_ms: int = 1000,
    ) -> None:
        """Initialize the manifest event handler

        :param pattern: a pattern that identifies the files whose
        events are of interest by matching their name
        :param ignore_patterns: a pattern that identifies the files whose
        events should be ignored
        :param ignore_directories: set to `True` to avoid directory events
        :param case_sensitive: set to `True` to require case sensitivity in
        resource names in order to match input patterns
        :param timeout_ms: maximum duration (in ms) of a call to the event
        loop prior to cancelling tasks
        """
        super().__init__(
            [pattern], ignore_patterns, ignore_directories, case_sensitive
        )  # type: ignore
        self._tracked_runs: t.Dict[int, Run] = {}
        self._tracked_jobs: t.Dict[_JobKey, JobEntity] = {}
        self._completed_jobs: t.Dict[_JobKey, JobEntity] = {}
        self._launcher: t.Optional[Launcher] = None
        self.job_manager: JobManager = JobManager(threading.RLock())
        self._launcher_map: t.Dict[str, t.Type[Launcher]] = {
            "slurm": SlurmLauncher,
            "pbs": PBSLauncher,
            "lsf": LSFLauncher,
            "local": LocalLauncher,
            "dragon": DragonLauncher,
        }
        self._collector_mgr = CollectorManager(timeout_ms)

    @property
    def tracked_jobs(self) -> t.Sequence[JobEntity]:
        """The collection of `JobEntity` that are actively being monitored

        :return: the collection
        """
        return list(self._tracked_jobs.values())

    def init_launcher(self, launcher: str) -> None:
        """Initialize the controller with a specific type of launcher.
        SmartSim currently supports Slurm, PBS(Pro), LSF, Dragon
        and local launching

        :param launcher: the name of the workload manager used by the experiment
        :raises ValueError: if a string is passed that is not
        a supported launcher
        :raises TypeError: if no launcher argument is provided.
        """
        if not launcher:
            raise TypeError("Must provide a 'launcher' argument")

        if launcher_type := self._launcher_map.get(launcher.lower(), None):
            self._launcher = launcher_type()
            return

        raise ValueError("Launcher type not supported: " + launcher)

    def init_job_manager(self) -> None:
        """Initialize the job manager instance"""
        if not self._launcher:
            raise TypeError("self._launcher must be initialized")

        self.job_manager.set_launcher(self._launcher)
        self.job_manager.start()

    def set_launcher(self, launcher_type: str) -> None:
        """Set the launcher for the experiment
        :param launcher_type: the name of the workload manager used by the experiment
        """
        self.init_launcher(launcher_type)

        if self._launcher is None:
            raise SmartSimError("Launcher init failed")

        self.job_manager.set_launcher(self._launcher)
        self.job_manager.start()

    def process_manifest(self, manifest_path: str) -> None:
        """Read the manifest for the experiment. Process the
        `RuntimeManifest` by updating the set of tracked jobs
        and registered collectors

        :param manifest_path: full path to the manifest file
        """
        try:
            # it is possible to read the manifest prior to a completed
            # write due to no access locking mechanism. log the issue
            # and continue. it will retry on the next event loop iteration
            manifest = RuntimeManifest.load_manifest(manifest_path)
            if not manifest:
                logger.debug("No manifest file exists")
                return
        except json.JSONDecodeError:
            logger.error(f"Malformed manifest encountered: {manifest_path}")
            return
        except ValueError:
            logger.error("Manifest content error", exc_info=True)
            return

        if self._launcher is None:
            self.set_launcher(manifest.launcher)

        if not self._launcher:
            raise SmartSimError(f"Unable to set launcher from {manifest_path}")

        # filter out previously added items
        runs = [run for run in manifest.runs if run.timestamp not in self._tracked_runs]

        # manifest is stored at <exp_dir>/.smartsim/telemetry/manifest.json
        exp_dir = pathlib.Path(manifest_path).parent.parent.parent

        for run in runs:
            for entity in run.flatten(
                filter_fn=lambda e: e.key not in self._tracked_jobs
            ):
                entity.path = str(exp_dir)

                # track everything coming in (managed and unmanaged)
                self._tracked_jobs[entity.key] = entity

                # register collectors for new entities as needed
                if entity.telemetry_on:
                    self._collector_mgr.register_collectors(entity)

                # persist a `start` event for each new entity in the manifest
                write_event(
                    run.timestamp,
                    entity.task_id,
                    entity.step_id,
                    entity.type,
                    "start",
                    pathlib.Path(entity.status_dir),
                )

                if entity.is_managed:
                    # Tell JobManager the task is unmanaged. This collects
                    # status updates but does not try to start a new copy
                    self.job_manager.add_job(
                        entity.name,
                        entity.step_id,
                        entity,
                        False,
                    )
                    # Tell the launcher it's managed so it doesn't attempt
                    # to look for a PID that may no longer exist
                    self._launcher.step_mapping.add(
                        entity.name, entity.step_id, "", True
                    )
            self._tracked_runs[run.timestamp] = run

    def on_modified(self, event: FileSystemEvent) -> None:
        """Event handler for when a file or directory is modified.

        :param event: event representing file/directory modification.
        """
        super().on_modified(event)
        logger.debug(f"Processing manifest modified @ {event.src_path}")
        self.process_manifest(event.src_path)

    def on_created(self, event: FileSystemEvent) -> None:
        """Event handler for when a file or directory is created.

        :param event: event representing file/directory creation.
        """
        super().on_created(event)
        logger.debug(f"processing manifest created @ {event.src_path}")
        self.process_manifest(event.src_path)

    async def _to_completed(
        self,
        timestamp: int,
        entity: JobEntity,
        step_info: StepInfo,
    ) -> None:
        """Move a monitored entity from the active to completed collection to
        stop monitoring for updates during timesteps.

        :param timestamp: current timestamp for event logging
        :param entity: running SmartSim Job
        :param step_info: `StepInfo` received when requesting a Job status update
        """
        # remember completed entities to ignore them after manifest updates
        inactive_entity = self._tracked_jobs.pop(entity.key)
        if entity.key not in self._completed_jobs:
            self._completed_jobs[entity.key] = inactive_entity

        # remove all the registered collectors for the completed entity
        await self._collector_mgr.remove(entity)

        job = self.job_manager[entity.name]
        self.job_manager.move_to_completed(job)

        status_clause = f"status: {step_info.status}"
        error_clause = f", error: {step_info.error}" if step_info.error else ""

        write_path = pathlib.Path(entity.status_dir)

        # persist a `stop` event for an entity that has completed
        write_event(
            timestamp,
            entity.task_id,
            entity.step_id,
            entity.type,
            "stop",
            write_path,
            detail=f"{status_clause}{error_clause}",
            return_code=map_return_code(step_info),
        )

    async def on_timestep(self, timestamp: int) -> None:
        """Called at polling frequency to request status updates on
        monitored entities

        :param timestamp: current timestamp for event logging
        """
        if not self._launcher:
            return

        await self._collector_mgr.collect()

        # ensure unmanaged jobs move out of tracked jobs list
        u_jobs = [job for job in self._tracked_jobs.values() if not job.is_managed]
        for job in u_jobs:
            job.check_completion_status()
            if job.is_complete:
                completed_entity = self._tracked_jobs.pop(job.key)
                self._completed_jobs[job.key] = completed_entity

        # consider not using name to avoid collisions
        m_jobs = [job for job in self._tracked_jobs.values() if job.is_managed]
        if names := {entity.name: entity for entity in m_jobs}:
            step_updates: t.List[t.Tuple[str, t.Optional[StepInfo]]] = []

            try:
                task_names = list(names.keys())
                updates = self._launcher.get_step_update(task_names)
                step_updates.extend(updates)
                logger.debug(f"Retrieved updates for: {task_names}")
            except Exception:
                logger.warning(f"Telemetry step updates failed for {names.keys()}")

            try:
                for step_name, step_info in step_updates:
                    if step_info and step_info.status in TERMINAL_STATUSES:
                        completed_entity = names[step_name]
                        await self._to_completed(timestamp, completed_entity, step_info)
            except Exception as ex:
                msg = f"An error occurred getting step updates on {names}"
                logger.error(msg, exc_info=ex)

    async def shutdown(self) -> None:
        """Release all resources owned by the `ManifestEventHandler`"""
        logger.debug(f"{type(self).__name__} shutting down...")
        await self._collector_mgr.shutdown()
        logger.debug(f"{type(self).__name__} shutdown complete...")


class TelemetryMonitorArgs:
    """Strongly typed entity to house logic for validating
    configuration passed to the telemetry monitor"""

    def __init__(
        self,
        exp_dir: str,
        frequency: int,
        cooldown: int,
        log_level: int = logging.DEBUG,
    ) -> None:
        """Initialize the instance with inputs and defaults

        :param exp_dir: root path to experiment outputs
        :param frequency: desired frequency of metric & status updates (in seconds)
        :param frequency: cooldown period (in seconds) before automatic shutdown
        :param log_level: log level to apply to python logging
        """
        self.exp_dir: str = exp_dir
        self.frequency: int = frequency  # freq in seconds
        self.cooldown: int = cooldown  # cooldown in seconds
        self.log_level: int = log_level
        self._validate()

    @property
    def min_frequency(self) -> int:
        """The minimum duration (in seconds) for the monitoring loop to wait
        between executions of the monitoring loop. Shorter frequencies may
        not allow the monitoring loop to complete. Adjusting the minimum frequency
        can result in inconsistent or missing outputs due to the telemetry
        monitor cancelling processes that exceed the allotted frequency."""
        return 1

    @property
    def max_frequency(self) -> int:
        """The maximum duration (in seconds) for the monitoring loop to wait
        between executions of the monitoring loop. Longer frequencies potentially
        keep the telemetry monitor alive unnecessarily."""
        return 600

    @property
    def min_cooldown(self) -> int:
        """The minimum allowed cooldown period that can be configured. Ensures
        the cooldown does not cause the telemetry monitor to shutdown prior to
        completing a single pass through the monitoring loop"""
        return min(self.frequency + 1, self.cooldown)

    @property
    def max_cooldown(self) -> int:
        """The maximum allowed cooldown period that can be configured. Ensures the
        telemetry monitor can automatically shutdown if not needed"""
        return self.max_frequency

    @property
    def cooldown_ms(self) -> int:
        """The duration of the time period (in ms) the telemetry monitor will
        wait for new resources to monitor before shutting down"""
        return self.cooldown * 1000

    @property
    def frequency_ms(self) -> int:
        """The desired frequency (in ms) of the telemetry monitor attempts
        to retrieve status updates and metrics"""
        return self.frequency * 1000

    def _check_exp_dir(self) -> None:
        """Validate the existence of the experiment directory"""
        if not pathlib.Path(self.exp_dir).exists():
            raise ValueError(f"Experiment directory cannot be found: {self.exp_dir}")

    def _check_frequency(self) -> None:
        """Validate the frequency input is in the range
        [`min_frequency`, `max_frequency`]"""
        if self.max_frequency >= self.frequency >= self.min_frequency:
            return

        freq_tpl = "Telemetry collection frequency must be in the range [{0}, {1}]"
        raise ValueError(freq_tpl.format(self.min_frequency, self.max_frequency))

    def _check_log_level(self) -> None:
        """Validate the frequency log level input. Uses standard python log levels"""
        if self.log_level not in [
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
        ]:
            raise ValueError(f"Invalid log_level supplied: {self.log_level}")

    def _validate(self) -> None:
        """Execute all validation functions"""
        self._check_exp_dir()
        self._check_frequency()
        self._check_log_level()


class TelemetryMonitor:
    """The telemetry monitor is a standalone process managed by SmartSim to perform
    long-term retrieval of experiment status updates and resource usage
    metrics. Note that a non-blocking driver script is likely to complete before
    the SmartSim entities complete. Also, the JobManager performs status updates
    only as long as the driver is running. This telemetry monitor entrypoint is
    started automatically when a SmartSim experiment calls the `start` method
    on resources. The entrypoint runs until it has no resources to monitor."""

    def __init__(self, telemetry_monitor_args: TelemetryMonitorArgs):
        """Initialize the telemetry monitor instance

        :param telemetry_monitor_args: configuration for the telemetry monitor
        """
        self._observer: BaseObserver = Observer()
        """an observer object that triggers the action handler"""
        self._args = telemetry_monitor_args
        """user-supplied arguments configuring telemetry monitor behavior"""
        self._experiment_dir = pathlib.Path(self._args.exp_dir)
        """path to the root directory where experiment outputs are written"""
        self._telemetry_path = self._experiment_dir / CONFIG.telemetry_subdir
        """path to the root directory where telemetry outputs are written"""
        self._manifest_path = self._telemetry_path / MANIFEST_FILENAME
        """path to the runtime manifest file"""
        self._action_handler: t.Optional[ManifestEventHandler] = None
        """an event listener holding action handlers for manifest on-change events"""

    def _can_shutdown(self) -> bool:
        """Determines if the telemetry monitor can perform shutdown. An
        automatic shutdown will occur if there are no active jobs being monitored.
        Managed jobs and databases are considered separately due to the way they
        are stored in the job manager

        :return: return True if capable of automatically shutting down
        """
        managed_jobs = (
            list(self._action_handler.job_manager.jobs.values())
            if self._action_handler
            else []
        )
        unmanaged_jobs = (
            list(self._action_handler.tracked_jobs) if self._action_handler else []
        )
        # get an individual count of databases for logging
        n_dbs: int = len(
            [
                job
                for job in managed_jobs + unmanaged_jobs
                if isinstance(job, JobEntity) and job.is_db
            ]
        )

        # if we have no jobs currently being monitored we can shutdown
        n_jobs = len(managed_jobs) + len(unmanaged_jobs) - n_dbs
        shutdown_ok = n_jobs + n_dbs == 0

        logger.debug(f"{n_jobs} active job(s), {n_dbs} active db(s)")
        return shutdown_ok

    async def monitor(self) -> None:
        """The main monitoring loop. Executes a busy wait and triggers
        telemetry collectors using frequency from constructor arguments.
        Continue monitoring until it satisfies automatic shutdown criteria."""
        elapsed: int = 0
        last_ts: int = get_ts_ms()
        shutdown_in_progress = False

        if self._action_handler is None:
            raise ValueError("The action handler must be initialized to monitor")

        # Event loop runs until the observer shuts down or
        # an automatic shutdown is started.
        while self._observer.is_alive() and not shutdown_in_progress:
            duration_ms = 0
            start_ts = get_ts_ms()
            await self._action_handler.on_timestep(start_ts)

            elapsed += start_ts - last_ts
            last_ts = start_ts

            # check if there are no jobs being monitored
            if self._can_shutdown():
                # cooldown period begins accumulating when no entities are monitored
                if elapsed >= self._args.cooldown_ms:
                    shutdown_in_progress = True
                    logger.info("Cooldown complete. Beginning shutdown")
                    await self._action_handler.shutdown()
                    logger.debug("Beginning file monitor shutdown")
                    self._observer.stop()  # type: ignore
                    logger.debug("Event loop shutdown complete")
                    break
            else:
                # reset cooldown any time jobs are running
                elapsed = 0

            # track time elapsed to execute metric collection
            duration_ms = get_ts_ms() - start_ts
            wait_ms = max(self._args.frequency_ms - duration_ms, 0)

            # delay next loop if collection time didn't exceed loop frequency
            wait_sec = wait_ms / 1000  # convert to seconds for sleep
            if elapsed > 0:
                completion_pct = elapsed / self._args.cooldown_ms * 100
                logger.info(f"Cooldown {completion_pct:.2f}% complete")
            logger.debug(f"Collection in {wait_sec:.2f}s")
            await asyncio.sleep(wait_sec)

        logger.info("Exiting telemetry monitor event loop")

    async def run(self) -> int:
        """Setup the monitoring entities and start the timer-based loop that
        will poll for telemetry data

        :return: return code for the process
        """
        logger.info("Executing telemetry monitor")
        logger.info(f"Polling frequency: {self._args.frequency}s")
        logger.info(f"Experiment directory: {self._experiment_dir}")
        logger.info(f"Telemetry output: {self._telemetry_path}")

        # Convert second-based inputs to milliseconds
        frequency_ms = int(self._args.frequency * 1000)

        # Create event handlers to trigger when target files are changed
        log_handler = LoggingEventHandler(logger)
        self._action_handler = ManifestEventHandler(
            str(MANIFEST_FILENAME),
            timeout_ms=frequency_ms,
            ignore_patterns=["*.out", "*.err"],
        )

        try:
            # The manifest may not exist when the telemetry monitor starts
            if self._manifest_path.exists():
                self._action_handler.process_manifest(str(self._manifest_path))

            # Add a handler to log file-system events
            self._observer.schedule(log_handler, self._telemetry_path)  # type:ignore
            # Add a handler to perform actions on file-system events
            self._observer.schedule(
                self._action_handler, self._telemetry_path
            )  # type:ignore
            self._observer.start()  # type: ignore

            # kick off the 'infinite' monitoring loop
            await self.monitor()
            return os.EX_OK
        except Exception as ex:
            logger.error(ex)
        finally:
            await self._action_handler.shutdown()
            self.cleanup()
            logger.info("Telemetry monitor shutdown complete")

        return os.EX_SOFTWARE

    def cleanup(self) -> None:
        """Perform cleanup for all allocated resources"""
        if self._observer is not None and self._observer.is_alive():
            logger.debug("Cleaning up manifest observer")
            self._observer.stop()  # type: ignore
            self._observer.join()



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from ....error import LauncherError
from ....log import get_logger
from ...utils.helpers import expand_exe_path
from ..util.shell import execute_cmd

logger = get_logger(__name__)


def sstat(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[str, str]:
    """Calls sstat with args

    :param args: List of command arguments
    :returns: Output and error of sstat
    """
    _, out, err = _execute_slurm_cmd("sstat", args, raise_on_err=raise_on_err)
    return out, err


def sacct(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[str, str]:
    """Calls sacct with args

    :param args: List of command arguments
    :returns: Output and error of sacct
    """
    _, out, err = _execute_slurm_cmd("sacct", args, raise_on_err=raise_on_err)
    return out, err


def salloc(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[str, str]:
    """Calls slurm salloc with args

    :param args: List of command arguments
    :returns: Output and error of salloc
    """
    _, out, err = _execute_slurm_cmd("salloc", args, raise_on_err=raise_on_err)
    return out, err


def sinfo(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[str, str]:
    """Calls slurm sinfo with args

    :param args: List of command arguments
    :returns: Output and error of sinfo
    """
    _, out, err = _execute_slurm_cmd("sinfo", args, raise_on_err=raise_on_err)
    return out, err


def scontrol(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[str, str]:
    """Calls slurm scontrol with args

    :param args: List of command arguments
    :returns: Output and error of sinfo
    """
    _, out, err = _execute_slurm_cmd("scontrol", args, raise_on_err=raise_on_err)
    return out, err


def scancel(args: t.List[str], *, raise_on_err: bool = False) -> t.Tuple[int, str, str]:
    """Calls slurm scancel with args.

    returncode is also supplied in this function.

    :param args: list of command arguments
    :return: output and error
    """
    return _execute_slurm_cmd("scancel", args, raise_on_err=raise_on_err)


def _find_slurm_command(cmd: str) -> str:
    try:
        full_cmd = expand_exe_path(cmd)
        return full_cmd
    except (TypeError, FileNotFoundError) as e:
        raise LauncherError(
            f"Slurm Launcher could not find path of {cmd} command"
        ) from e


def _execute_slurm_cmd(
    command: str, args: t.List[str], raise_on_err: bool = False
) -> t.Tuple[int, str, str]:
    cmd_exe = _find_slurm_command(command)
    cmd = [cmd_exe] + args
    returncode, out, error = execute_cmd(cmd)
    if returncode != 0:
        msg = f"An error occurred while calling {command}: {error}"
        if raise_on_err:
            raise LauncherError(msg)
        logger.error(msg)
    return returncode, out, error



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import typing as t
from shlex import split as sh_split

from ....error import AllocationError
from ....log import get_logger
from ....settings import RunSettings, SbatchSettings, Singularity, SrunSettings
from .step import Step

logger = get_logger(__name__)


class SbatchStep(Step):
    def __init__(self, name: str, cwd: str, batch_settings: SbatchSettings) -> None:
        """Initialize a Slurm Sbatch step

        :param name: name of the entity to launch
        :param cwd: path to launch dir
        :param batch_settings: batch settings for entity
        """
        super().__init__(name, cwd, batch_settings)
        self.step_cmds: t.List[t.List[str]] = []
        self.managed = True
        self.batch_settings = batch_settings

    def get_launch_cmd(self) -> t.List[str]:
        """Get the launch command for the batch

        :return: launch command for the batch
        """
        script = self._write_script()
        return [self.batch_settings.batch_cmd, "--parsable", script]

    def add_to_batch(self, step: Step) -> None:
        """Add a job step to this batch

        :param step: a job step instance e.g. SrunStep
        """
        launch_cmd = ["cd", step.cwd, ";"]
        launch_cmd += step.get_launch_cmd()
        self.step_cmds.append(launch_cmd)
        logger.debug(f"Added step command to batch for {step.name}")

    def _write_script(self) -> str:
        """Write the batch script

        :return: batch script path after writing
        """
        batch_script = self.get_step_file(ending=".sh")
        output, error = self.get_output_files()
        with open(batch_script, "w", encoding="utf-8") as script_file:
            script_file.write("#!/bin/bash\n\n")
            script_file.write(f"#SBATCH --output={output}\n")
            script_file.write(f"#SBATCH --error={error}\n")
            script_file.write(f"#SBATCH --job-name={self.name}\n")

            # add additional sbatch options
            for opt in self.batch_settings.format_batch_args():
                script_file.write(f"#SBATCH {opt}\n")

            for cmd in self.batch_settings.preamble:
                script_file.write(f"{cmd}\n")

            for i, step_cmd in enumerate(self.step_cmds):
                script_file.write("\n")
                script_file.write(f"{' '.join((step_cmd))} &\n")
                if i == len(self.step_cmds) - 1:
                    script_file.write("\n")
                    script_file.write("wait\n")
        return batch_script


class SrunStep(Step):
    def __init__(self, name: str, cwd: str, run_settings: SrunSettings) -> None:
        """Initialize a srun job step

        :param name: name of the entity to be launched
        :param cwd: path to launch dir
        :param run_settings: run settings for entity
        """
        super().__init__(name, cwd, run_settings)
        self.alloc: t.Optional[str] = None
        self.managed = True
        self.run_settings = run_settings
        if not self.run_settings.in_batch:
            self._set_alloc()

    def get_launch_cmd(self) -> t.List[str]:
        """Get the command to launch this step

        :return: launch command
        """
        srun = self.run_settings.run_command
        if not srun:
            raise ValueError("No srun command found in PATH")

        output, error = self.get_output_files()

        srun_cmd = [srun, "--output", output, "--error", error, "--job-name", self.name]
        compound_env: t.Set[str] = set()

        if self.alloc:
            srun_cmd += ["--jobid", str(self.alloc)]

        if self.run_settings.env_vars:
            env_vars, csv_env_vars = self.run_settings.format_comma_sep_env_vars()

            if len(env_vars) > 0:
                srun_cmd += ["--export", f"ALL,{env_vars}"]

            if csv_env_vars:
                compound_env = compound_env.union(csv_env_vars)

        srun_cmd += self.run_settings.format_run_args()

        if self.run_settings.colocated_db_settings:
            # Replace the command with the entrypoint wrapper script
            bash = shutil.which("bash")
            if not bash:
                raise RuntimeError("Could not find bash in PATH")
            launch_script_path = self.get_colocated_launch_script()
            srun_cmd += [bash, launch_script_path]

        if isinstance(self.run_settings.container, Singularity):
            # pylint: disable-next=protected-access
            srun_cmd += self.run_settings.container._container_cmds(self.cwd)

        if compound_env:
            srun_cmd = ["env"] + list(compound_env) + srun_cmd

        srun_cmd += self._build_exe()

        return srun_cmd

    def _set_alloc(self) -> None:
        """Set the id of the allocation

        :raises AllocationError: allocation not listed or found
        """
        if self.run_settings.alloc:
            self.alloc = self.run_settings.alloc
        else:
            if "SLURM_JOB_ID" in os.environ:
                self.alloc = os.environ["SLURM_JOB_ID"]
                logger.debug(
                    f"Running on allocation {self.alloc} gleaned from user environment"
                )
            else:
                raise AllocationError(
                    "No allocation specified or found and not running in batch"
                )

    def _get_mpmd(self) -> t.List[RunSettings]:
        """Temporary convenience function to return a typed list
        of attached RunSettings
        """
        return self.run_settings.mpmd

    @staticmethod
    def _get_exe_args_list(run_setting: RunSettings) -> t.List[str]:
        """Convenience function to encapsulate checking the
        runsettings.exe_args type to always return a list
        """
        exe_args = run_setting.exe_args
        args: t.List[str] = exe_args if isinstance(exe_args, list) else [exe_args]
        return args

    def _build_exe(self) -> t.List[str]:
        """Build the executable for this step

        :return: executable list
        """
        if self._get_mpmd():
            return self._make_mpmd()

        exe = self.run_settings.exe
        args = self._get_exe_args_list(self.run_settings)
        return exe + args

    def _make_mpmd(self) -> t.List[str]:
        """Build Slurm multi-prog (MPMD) executable"""
        exe = self.run_settings.exe
        args = self._get_exe_args_list(self.run_settings)
        cmd = exe + args

        compound_env_vars = []
        for mpmd_rs in self._get_mpmd():
            cmd += [" : "]
            cmd += mpmd_rs.format_run_args()
            cmd += ["--job-name", self.name]

            if isinstance(mpmd_rs, SrunSettings):
                (env_var_str, csv_env_vars) = mpmd_rs.format_comma_sep_env_vars()
                if len(env_var_str) > 0:
                    cmd += ["--export", f"ALL,{env_var_str}"]
                if csv_env_vars:
                    compound_env_vars.extend(csv_env_vars)
            cmd += mpmd_rs.exe
            cmd += self._get_exe_args_list(mpmd_rs)

        cmd = sh_split(" ".join(cmd))
        return cmd



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
import typing as t
from functools import lru_cache
from pathlib import Path

import psutil

from ...error import SSConfigError
from ..utils.helpers import expand_exe_path

# Configuration Values
#
# These values can be set through environment variables to
# override the default behavior of SmartSim.
#
# RAI_PATH
#   - Path to the RAI shared library
#   - Default: /smartsim/smartsim/_core/lib/redisai.so
#
# REDIS_CONF
#   - Path to the redis.conf file
#   - Default: /SmartSim/smartsim/_core/config/redis.conf
#
# REDIS_PATH
#   - Path to the redis-server executable
#   - Default: /SmartSim/smartsim/_core/bin/redis-server
#
# REDIS_CLI_PATH
#   - Path to the redis-cli executable
#   - Default: /SmartSim/smartsim/_core/bin/redis-cli
#
# SMARTSIM_LOG_LEVEL
#   - Log level for SmartSim
#   - Default: info
#
# SMARTSIM_JM_INTERVAL
#   - polling interval for communication with scheduler
#   - default: 10 seconds
#


# Testing Configuration Values
#
# SMARTSIM_TEST_INTERFACE
#  - Network interface to use for testing
#  - Default: auto-detected
#
# SMARTSIM_TEST_LAUNCHER
#  - type of launcher to use for testing
#  - Default: Local
#
# SMARTSIM_TEST_DEVICE
#  - CPU or GPU for model serving tests
#  - Default: CPU
#
# SMARTSIM_TEST_NUM_GPUS
#  - Number of GPUs on the host for testing
#  - Defaults: 1
#
# SMARTSIM_TEST_ACCOUNT
#  - Account used to run full launcher test suite on external systems
#  - Default: None


# pylint: disable-next=too-many-public-methods
class Config:
    def __init__(self) -> None:
        # SmartSim/smartsim/_core
        self.core_path = Path(os.path.abspath(__file__)).parent.parent

        dependency_path = os.environ.get("SMARTSIM_DEP_INSTALL_PATH", self.core_path)

        self.lib_path = Path(dependency_path, "lib").resolve()
        self.bin_path = Path(dependency_path, "bin").resolve()
        self.conf_path = Path(dependency_path, "config", "redis.conf")
        self.conf_dir = Path(self.core_path, "config")

    @property
    def redisai(self) -> str:
        rai_path = self.lib_path / "redisai.so"
        redisai = Path(os.environ.get("RAI_PATH", rai_path)).resolve()
        if not redisai.is_file():
            raise SSConfigError(
                "RedisAI dependency not found. Build with `smart` cli "
                "or specify RAI_PATH"
            )
        return str(redisai)

    @property
    def database_conf(self) -> str:
        conf = Path(os.environ.get("REDIS_CONF", self.conf_path)).resolve()
        if not conf.is_file():
            raise SSConfigError(
                "Database configuration file at REDIS_CONF could not be found"
            )
        return str(conf)

    @property
    def database_exe(self) -> str:
        try:
            database_exe = next(self.bin_path.glob("*-server"))
            database = Path(os.environ.get("REDIS_PATH", database_exe)).resolve()
            exe = expand_exe_path(str(database))
            return exe
        except (TypeError, FileNotFoundError) as e:
            raise SSConfigError(
                "Specified database binary at REDIS_PATH could not be used"
            ) from e

    @property
    def database_cli(self) -> str:
        try:
            redis_cli_exe = next(self.bin_path.glob("*-cli"))
            redis_cli = Path(os.environ.get("REDIS_CLI_PATH", redis_cli_exe)).resolve()
            exe = expand_exe_path(str(redis_cli))
            return exe
        except (TypeError, FileNotFoundError) as e:
            raise SSConfigError(
                "Specified Redis binary at REDIS_CLI_PATH could not be used"
            ) from e

    @property
    def database_file_parse_trials(self) -> int:
        return int(os.getenv("SMARTSIM_DB_FILE_PARSE_TRIALS", "10"))

    @property
    def database_file_parse_interval(self) -> int:
        return int(os.getenv("SMARTSIM_DB_FILE_PARSE_INTERVAL", "2"))

    @property
    def dragon_dotenv(self) -> Path:
        """Returns the path to a .env file containing dragon environment variables"""
        return self.conf_dir / "dragon" / ".env"

    @property
    def dragon_server_path(self) -> t.Optional[str]:
        return os.getenv(
            "SMARTSIM_DRAGON_SERVER_PATH",
            os.getenv("SMARTSIM_DRAGON_SERVER_PATH_EXP", None),
        )

    @property
    def dragon_server_timeout(self) -> int:
        return int(os.getenv("SMARTSIM_DRAGON_TIMEOUT", "30000"))

    @property
    def dragon_server_startup_timeout(self) -> int:
        return int(os.getenv("SMARTSIM_DRAGON_STARTUP_TIMEOUT", "300000"))

    @property
    def dragon_transport(self) -> str:
        return os.getenv("SMARTSIM_DRAGON_TRANSPORT", "hsta")

    @property
    def log_level(self) -> str:
        return os.environ.get("SMARTSIM_LOG_LEVEL", "info")

    @property
    def jm_interval(self) -> int:
        return int(os.environ.get("SMARTSIM_JM_INTERVAL") or 10)

    @property
    def wlm_trials(self) -> int:
        return int(os.environ.get("SMARTSIM_WLM_TRIALS") or 10)

    @property
    def test_launcher(self) -> str:  # pragma: no cover
        return os.environ.get("SMARTSIM_TEST_LAUNCHER", "local")

    @property
    def test_device(self) -> str:  # pragma: no cover
        return os.environ.get("SMARTSIM_TEST_DEVICE", "CPU")

    @property
    def test_num_gpus(self) -> int:  # pragma: no cover
        return int(os.environ.get("SMARTSIM_TEST_NUM_GPUS") or 1)

    @property
    def test_ports(self) -> t.Sequence[int]:  # pragma: no cover
        min_required_ports = 25
        first_port = int(os.environ.get("SMARTSIM_TEST_PORT", 6780))
        num_ports = max(
            int(os.environ.get("SMARTSIM_TEST_NUM_PORTS", min_required_ports)),
            min_required_ports,
        )
        return range(first_port, first_port + num_ports)

    @property
    def test_batch_resources(self) -> t.Dict[t.Any, t.Any]:  # pragma: no cover
        resource_str = os.environ.get("SMARTSIM_TEST_BATCH_RESOURCES", "{}")
        resources = json.loads(resource_str)
        if not isinstance(resources, dict):
            raise TypeError(
                (
                    "SMARTSIM_TEST_BATCH_RESOURCES was not interpreted as a "
                    "dictionary, check to make sure that it is a valid "
                    f"JSON string: {resource_str}"
                )
            )
        return resources

    @property
    def test_interface(self) -> t.List[str]:  # pragma: no cover
        if interfaces_cfg := os.environ.get("SMARTSIM_TEST_INTERFACE", None):
            return interfaces_cfg.split(",")

        # try to pick a sensible one
        net_if_addrs = psutil.net_if_addrs()
        if "ipogif0" in net_if_addrs:
            return ["ipogif0"]
        if "hsn0" in net_if_addrs:
            return [
                net_if_addr
                for net_if_addr in net_if_addrs
                if net_if_addr.startswith("hsn")
            ]
        if "ib0" in net_if_addrs:
            return ["ib0"]
        # default to aries network
        return ["lo"]

    @property
    def test_account(self) -> t.Optional[str]:  # pragma: no cover
        # no account by default
        return os.environ.get("SMARTSIM_TEST_ACCOUNT", None)

    @property
    def test_mpi(self) -> bool:  # pragma: no cover
        # By default, test MPI app if it compiles
        return int(os.environ.get("SMARTSIM_TEST_MPI", "1")) > 0

    @property
    def telemetry_frequency(self) -> int:
        return int(os.environ.get("SMARTSIM_TELEMETRY_FREQUENCY", 5))

    @property
    def telemetry_enabled(self) -> bool:
        return int(os.environ.get("SMARTSIM_FLAG_TELEMETRY", "1")) > 0

    @property
    def telemetry_cooldown(self) -> int:
        return int(os.environ.get("SMARTSIM_TELEMETRY_COOLDOWN", 90))

    @property
    def telemetry_subdir(self) -> str:
        return ".smartsim/telemetry"

    @property
    def dragon_default_subdir(self) -> str:
        return ".smartsim/dragon"

    @property
    def dragon_log_filename(self) -> str:
        return "dragon_config.log"

    @property
    def smartsim_key_path(self) -> str:
        """Path to a root directory used for persistence of key files. Default
        value `$HOME/.smartsim/keys`. User-overrideable by setting the environment
        variable `SMARTSIM_KEY_PATH`.

        :returns: The configured key path.
        """
        default_path = Path.home() / ".smartsim" / "keys"
        return os.environ.get("SMARTSIM_KEY_PATH", str(default_path))

    @property
    def dragon_pin(self) -> str:
        return "0.9"


@lru_cache(maxsize=128, typed=False)
def get_config() -> Config:
    # wrap into a function with a cached result
    return Config()



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from .entity import SmartSimEntity

if t.TYPE_CHECKING:
    # pylint: disable-next=unused-import
    import smartsim

_T = t.TypeVar("_T", bound=SmartSimEntity)
# Old style pyint from TF 2.6.x does not know about pep484 style ``TypeVar`` names
# pylint: disable-next=invalid-name
_T_co = t.TypeVar("_T_co", bound=SmartSimEntity, covariant=True)


class EntitySequence(t.Generic[_T_co]):
    """Abstract class for containers for SmartSimEntities"""

    def __init__(self, name: str, path: str, **kwargs: t.Any) -> None:
        self.name: str = name
        self.path: str = path

        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
        # WARNING: This class cannot be made truly covariant until the
        #          following properties are made read-only. It is currently
        #          designed for in-house type checking only!!
        #
        # Despite the fact that these properties are type hinted as
        # ``Sequence``s, the underlying types must remain ``list``s as that is
        # what subclasses are expecting when implementing their
        # ``_initialize_entities`` methods.
        #
        # I'm leaving it "as is" for now as to not introduce a potential API
        # break in case any users subclassed the invariant version of this
        # class (``EntityList``), but a "proper" solution would be to turn
        # ``EntitySequence``/``EntityList`` into proper ``abc.ABC``s and have
        # the properties we expect to be initialized represented as abstract
        # properties. An additional benefit of this solution is would be that
        # users could actually initialize their entities in the ``__init__``
        # method, and it would remove the need for the cumbersome and
        # un-type-hint-able ``_initialize_entities`` method by returning all
        # object construction into the class' constructor.
        # ---------------------------------------------------------------------
        #
        self.entities: t.Sequence[_T_co] = []
        self._db_models: t.Sequence["smartsim.entity.DBModel"] = []
        self._db_scripts: t.Sequence["smartsim.entity.DBScript"] = []
        #
        # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

        self._initialize_entities(**kwargs)

    def _initialize_entities(self, **kwargs: t.Any) -> None:
        """Initialize the SmartSimEntity objects in the container"""
        raise NotImplementedError

    @property
    def db_models(self) -> t.Iterable["smartsim.entity.DBModel"]:
        """Return an immutable collection of attached models"""
        return (model for model in self._db_models)

    @property
    def db_scripts(self) -> t.Iterable["smartsim.entity.DBScript"]:
        """Return an immutable collection of attached scripts"""
        return (script for script in self._db_scripts)

    @property
    def batch(self) -> bool:
        """Property indicating whether or not the entity sequence should be
        launched as a batch job

        :return: ``True`` if entity sequence should be launched as a batch job,
                 ``False`` if the members will be launched individually.
        """
        # pylint: disable-next=no-member
        return hasattr(self, "batch_settings") and self.batch_settings

    @property
    def type(self) -> str:
        """Return the name of the class"""
        return type(self).__name__

    def set_path(self, new_path: str) -> None:
        self.path = new_path
        for entity in self.entities:
            entity.path = new_path

    def __getitem__(self, name: str) -> t.Optional[_T_co]:
        for entity in self.entities:
            if entity.name == name:
                return entity
        return None

    def __iter__(self) -> t.Iterator[_T_co]:
        for entity in self.entities:
            yield entity

    def __len__(self) -> int:
        return len(self.entities)


class EntityList(EntitySequence[_T]):
    """An invariant subclass of an ``EntitySequence`` with mutable containers"""

    def __init__(self, name: str, path: str, **kwargs: t.Any) -> None:
        super().__init__(name, path, **kwargs)
        # Change container types to be invariant ``list``s
        self.entities: t.List[_T] = list(self.entities)
        self._db_models: t.List["smartsim.entity.DBModel"] = list(self._db_models)
        self._db_scripts: t.List["smartsim.entity.DBScript"] = list(self._db_scripts)

    def _initialize_entities(self, **kwargs: t.Any) -> None:
        """Initialize the SmartSimEntity objects in the container"""
        # Need to identically re-define this "abstract method" or pylint
        # complains that we are trying to define a concrete implementation of
        # an abstract class despite the fact that we want this class to also be
        # abstract.  All the more reason to turn both of these classes into
        # ``abc.ABC``s in my opinion.
        raise NotImplementedError
